# Manage Trac Daemon, Accounts and Interoperations.
from stuphos.runtime.facilities import Facility
from stuphmud.server.player.shell import PromptFor
from stuphmud.server.player import ACMDLN, Option
from ..etc.hashlib import md5
from ..etc import OpenFile, FileNotFoundException
from stuphos import getSection

# Dependency:
try: from tracacctmgrrpc.client import JsonRpc
except ImportError:
    JsonRpc = NotImplementedError('tracacctmgrrpc.client.JsonRpc')

# import re

##    class HTPasswd:
##        class UserAlreadyExists(Exception):
##            pass
##
##        PATTERN = re.compile('(\w+)\:(\w*)\:(\w*)').match
##
##        def __init__(self, filename, default_realm):
##            self.filename = filename
##            self.default_realm = default_realm
##            self.byRealm = dict() # composed?
##
##        def __repr__(self):
##            return '<%s: %s>' % (self.__class__.__name__, self.filename)
##
##        def getRealm(self, realm = None):
##            return self.default_realm if realm is None else realm
##
##        def clear(self):
##            self.byRealm.clear()
##
##        def parse(self, source):
##            for line in source:
##                p = self.PATTERN(line)
##                if p is not None:
##                    (user, realm, password) = p.groups()
##                    yield (realm, user, password)
##
##        def __iter__(self):
##            for (realm, entries) in self.byRealm.iteritems():
##                for (user, password) in entries.iteritems():
##                    yield (realm, user, password)
##
##        def write(self, stream):
##            for (realm, user, password) in self:
##                print >> stream, '%s:%s:%s' % (user, realm, password)
##
##        def load(self):
##            try:
##                with OpenFile(self.filename) as i:
##                    for (realm, user, password) in self.parse(i):
##                        self.byRealm.setdefault(realm, {})[user] = password
##
##            except FileNotFoundException:
##                pass
##
##        def save(self, transaction = None):
##            # Todo: implement transaction mode as something that can do backup/journal.
##            # assert not transaction, NotImplementedError
##            self.write(open(self.filename, 'w'))
##
##        def getAllRealms(self):
##            return self.byRealm.iterkeys()
##        def getUsersForRealm(self, realm = None):
##            return self.byRealm[self.getRealm(realm)].iteritems()
##
##        def crypt(self, username, password, realm = None):
##            digest = md5(':'.join((username, self.getRealm(realm), password)))
##            return digest.hexdigest()
##
##        def addUser(self, username, password, realm = None, allowChange = False, saveChanges = False, cleartext = True, transaction = None):
##            realm = self.getRealm(realm)
##
##            try: r = self.byRealm[realm]
##            except KeyError:
##                r = self.byRealm[realm] = dict()
##
##            # If, so just change it!
##            if not allowChange and username in r:
##                raise HTPasswd.UserAlreadyExists(username)
##
##            if cleartext:
##                password = self.crypt(username, password, realm)
##
##            r[username] = password
##
##            if saveChanges:
##                self.save(transaction = transaction)
##
##        class ByUser:
##            def __init__(self, username, password, realm = None):
##                self.username = username
##                self.password = password
##                self.realm = realm
##
##            def __repr__(self):
##                return 'user: %s' % self.username
##
##        def getUser(self, username, realm = None):
##            return self.ByUser(username, self.byRealm[self.getRealm(realm)][username])
##        def getAllUsers(self):
##            for r in self.getAllRealms():
##                for (username, password) in self.getUsersForRealm(r):
##                    yield self.ByUser(username, password, realm = r)
##
##        def getAllUsersByName(self, thisUser):
##            for r in self.getAllRealms():
##                for (username, password) in self.getUsersForRealm(r):
##                    if username == thisUser:
##                        yield self.ByUser(username, password, realm = r)
##
##        def __iadd__(self, user):
##            if isinstance(user, self.ByUser):
##                username = user.username
##                password = user.password
##            else:
##                (username, password) = user
##
##            self.addUser(username, password)
##            return self
##
##        def __enter__(self):
##            self.load()
##            return self
##
##        def __exit__(self, etype = None, value = None, tb = None):
##            if value is not None:
##                self.save()
##
##
##    DEFAULT_REALM = 'trac' # 'stuph'
##    subdaemonMgrApi = runtime.SubDaemon.Manager.API

DEFAULT_REALM = 'trac' # 'stuph'
DEFAULT_HOSTNAME = 'localhost'

class TracSite(Facility):
    NAME = 'Trac::Site'

    class Manager(Facility.Manager):
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR
        VERB_NAME = 'trac-*site-manage'

        ##    def do_users(self, peer, cmd, argstr):
        ##        ts = TracSite.get()
        ##        if ts is not None:
        ##            @Showing(peer, 'TracSite Users')
        ##            def showUsers():
        ##                ht = ts.getHT()
        ##                for realm in ht.getAllRealms():
        ##                    yield '  %s:' % realm
        ##                    yield ''
        ##
        ##                    # Todo: columnize?
        ##                    for (user, passwd) in ht.getUsersForRealm(realm):
        ##                        yield '    %s' % user
        ##
        ##                    yield ''
        ##
        ##    def do_adduser(self, peer, cmd, argstr):
        ##        pass
        ##    def do_setpasswd(self, peer, cmd, argstr):
        ##        pass

    def __init__(self, trac = None):
        '''
        [TracSite]
        realm = trac
        port = 2184
        path = /login/rpc
        admin-username = acctmgr
        admin-password = abcdef
        greylist = op,admin,authenticated,anonymous,developers,webmasters
        greylist-file = %(etcdir)s/tracsite.greylist
        '''

        if trac is None:
            trac = getSection('TracSite')

        self.realm = trac.get('realm', DEFAULT_REALM)
        self.hostname = trac.get('hostname', DEFAULT_HOSTNAME)
        self.port = trac.get('port')
        self.path = trac.get('path')
        self.admin_username = trac.get('admin-username')
        self.admin_password = trac.get('admin-password')

    @property
    def rpc(self):
        try: return self._acctmgr_rpc
        except AttributeError:
            assert not isinstance(JsonRpc, Exception), JsonRpc
            rpc = self._acctmgr_rpc = JsonRpc(self.hostname,
                                              self.port,
                                              self.path,
                                              self.admin_username,
                                              self.admin_password,
                                              self.realm)

            return rpc

    ##    def __init__(self, trac = None):
    ##        '''
    ##        [TracSite]
    ##        realm = trac
    ##        passwdfile = %(etcdir)s/tracsite/conf/.passwd
    ##        subdaemon = trac-site
    ##        backupmode = full-journal, rotation
    ##        greylist = op,admin,authenticated,anonymous,developers,webmasters
    ##        greylist-file = %(etcdir)s/tracsite.greylist
    ##        '''
    ##
    ##        if trac is None:
    ##            trac = getSection('TracSite')
    ##
    ##        self.realm = trac.get('realm', DEFAULT_REALM)
    ##        self.passwdfile = trac.get('password-file')
    ##        self.subdaemon = trac.get('subdaemon-name')
    ##        self.backupmode = trac.get('backup-mode')
    ##        assert self.passwdfile
    ##
    ##    def getHT(self):
    ##        return HTPasswd(self.passwdfile, self.realm)
    ##
    ##    def getTransactionMode(self):
    ##        return self.backupmode
    ##
    ##    def ReloadDaemon(self):
    ##        # Todo: start-up, shutdown should be factored into the
    ##        # subdaemon manager entirely.  The idea is to prevent
    ##        # too many restarts in time...  Basically, we can't
    ##        # control how long it takes for tracd to start up, so
    ##        # we should keep track of the last time we booted it,
    ##        # and if it's been only a short time (like a minute?),
    ##        # then delay the reboot, and also tell the user how long
    ##        # we think it might take for the account to be available.
    ##        if self.subdaemon:
    ##            api = runtime[subdaemonMgrApi]
    ##            if api is not None:
    ##                # Reboot!
    ##                api.restartDaemon(self.subdaemon)

    class UserAlreadyExists(Exception):
        pass

    def addUser(self, playerName, password, peer = None):
        # Todo: interpret (JsonObject) response errors (for UserAlreadyExists)
        return self.rpc.addUser(playerName, password) == 0

        ##    with self.getHT() as ht:
        ##        ht.addUser(playerName, password, saveChanges = True,
        ##                   transaction = self.getTransactionMode())
        ##
        ##    # also, this requires rebooting the trac server subdaemon.
        ##    # todo: research authentication when using apache deployment
        ##    # todo: xmlrpc interface for reloading passwords? (custom plugin?)
        ##    self.ReloadDaemon()

        # Also, send confirmation email/mudmail?
        # Todo: use player.messaging facility

    class NewUserRequest:
        # Connect new-user request of management facility to player interactive elements.

        class PasswordEntry(PromptFor):
            # Todo: toggle echo..
            def __init__(self, request = None, confirming = None):
                if request is None:
                    request = confirming.request

                self.request = request
                self.ask = confirming

                method = self.askUserPassword if confirming is None else self.confirmUserPassword
                PromptFor.__init__(self, request.peer, method,
                                   message = self.getPromptForPasswordMessage(' E' if confirming is None else 'Re-e'),
                                   write_mode = True, compact_mode = True)

            def regenerate(self, peer):
                self.__class__(confirming = self)

            def getPromptForPasswordMessage(self, prefix):
                return '%snter password for [%s]: ' % (prefix, self.request.playerName)

            def askUserPassword(self, prompt, peer, password):
                if password:
                    self.password = password
                    return True # Regenerate

            def confirmUserPassword(self, prompt, peer, password):
                if password == self.ask.password:
                    userName = self.request.getUsername()

                    try: self.request.tracSite.addUser(userName, password, peer = peer)
                    # except HTPasswd.UserAlreadyExists, e:
                    except self.UserAlreadyExists as e:
                        print('&rUser already exists: %s&N' % e, file=peer)
                        print('&NUse &wrequest-trac-user&N &C--&Nchange&C-&Npassword', file=peer)
                    else:
                        print('User %r added to trac site.' % userName, file=peer)
                else:
                    print('Passwords do not match.  Try again.', file=peer)

        def __init__(self, tracSite, peer, playerName):
            self.tracSite = tracSite
            self.playerName = playerName
            self.peer = peer

        def getUsername(self):
            # Any reason to strip/validate?
            return self.playerName.lower()

        def userExists(self):
            for user in self.tracSite.getHT() \
                            .getAllUsersByName \
                                (self.getUsername()):

                return True # what about realm?

            return False

        def __call__(self):
            return self.PasswordEntry(self)

    def getPlayerUsername(self, player):
        return player.name

    def requestNewUser(self, peer, playerName):
        request = self.NewUserRequest(self, peer, playerName)
        return request()

        # Ignore this for right now because it means loading ht twice!
        ##    if not request.userExists():
        ##        return request()

@ACMDLN('request-trac-user*', Option('--action', default = 'new-user'),

                              Option('--change-password', action = 'store_const',
                                     const = 'change-password', dest = 'action'),

                              Option('--whoami', action = 'store_const',
                                     const = 'whoami', dest = 'action'),

                              Option('--new-user', action = 'store_const',
                                     const = 'new-user', dest = 'action'))

def doRequestTracUser(player, command):
    ts = TracSite.get()
    if ts is not None:
        a = player.avatar
        if a is not None: # and not a.npc:
            action = command.options.action

            if action == 'change-password':
                pass
            elif action == 'whoami':
                pass
            elif action == 'new-user':
                # And, delegate permissions to the facility?
                # (like, only allow immortals to create requests i this way)
                playerName = ts.getPlayerUsername(a)
                if ts.requestNewUser(player, playerName) is None:
                    print('User %r already exists in trac site.' % playerName, file=player)

            return True

# Load this as a facility/management services.
##    try: del runtime[TracSite.NAME]
##    except AttributeError: pass
##
##    runtime.Trac.Site(TracSite, {'password-file': 'passwd'})



# MUD Runtime -- Object Entities.
#
from .. import Concretion, declareEventController, EventController

# from types import DictionaryType
DictionaryType = dict
# from new import classobj as newClassObject
from types import new_class

def populateNamespace(data):
    def doNamespacePopulate(ns):
        ns.update(data)
    return doNamespacePopulate

def newClassObject(name, bases, values):
    return new_class(name, bases, exec_body = populateNamespace(values))

# Overall identity constant.
class Undefined(object):
    def __repr__(self):
        return self.__class__.__name__
    __str__ = __unicode__ = __repr__

Undefined = Undefined()

# Design Pattern.
class Singleton(object):
    # This should be further specialized into 'Event' for runtime constructs.
    class Meta(type):
        def __new__(self, name, bases, values):
            cls = type.__new__(self, name, bases, values)
            if Singleton in bases:
                return cls

            ##    postvalues = {}
            ##    for name in values.keys():
            ##        if name in ['__module__', '__doc__']:
            ##            postvalues[name] = values[name]
            ##            del values[name]
            ##
            ##    inst = cls(name, **values)
            ##    inst.__dict__.update(postvalues)
            ##    return inst

            return cls(name, **values)

# Identity.
class Object: # (object):
    class _Meta:
        Attributes = []

        def __init__(self, *attributes, **kwd):
            self.Attributes = list(attributes) + list(kwd.items())

        @staticmethod
        def formatAttribute(instance, a, default = Undefined):
            def getAttribute(name):
                if callable(name):
                    return name(instance)

                if name.endswith('()'):
                    v = getattr(instance._Meta, name[:-2], default)
                    if callable(v):
                        return v(instance)
                else:
                    return getattr(instance, name, default)

                return Undefined

            if type(a) in (list, tuple):
                if len(a) == 2:
                    return '%s = %r' % (a[0], getAttribute(a[1]))
                if len(a) == 3:
                    return '%s = %r' % (a[0], getAttribute(a[1], a[2]))

            elif type(a) is str:
                return '%s = %r' % (a, getAttribute(a))

        @staticmethod
        def className(instance):
            return instance.__object_name__()

        @classmethod
        def instanceRepr(self, instance):
            meta = instance._Meta
            attribs = ', '.join(meta.formatAttribute(instance, a) for \
                                a in meta.Attributes)
            if attribs:
                return '<%s %s>' % (meta.className(instance), attribs)

            return '<%s>' % meta.className(instance)

    def __init__(self, name = Undefined):
        if name is not Undefined:
            self.__name = name

    def __repr__(self):
        return self._Meta.instanceRepr(self)
    def __str__(self):
        return self.__repr__()

    # This should go in the Meta.
    def __object_name__(self):
        try: return self.__name
        except AttributeError:
            return self.__class__.__name__

    @classmethod
    def _instanceOf(self, other):
        return isinstance(other, self)

        ##    try: return issubclass(other.__class__, self)
        ##    except AttributeError:
        ##        return False

from .lookup import LookupObject
LookupClassObject = LookupObject

class Synthetic(Object, dict): # todo: determine if adding dict base type is stable
    class _Meta(Object._Meta):
        Attributes = Object._Meta.Attributes + ['members()']

        @staticmethod
        def members(instance):
            return ', '.join(map(str, list(instance.__dict__.keys())))

    def __init__(self, dict = None, **values):
        if not isinstance(dict, DictionaryType):
            assert dict is None
            dict = values
        else:
            dict.update(values)

        self.__dict__ = dict

    # todo: FromStructure and ToStructure methods like in WRLC

class Namespace(Synthetic):
    # XXX :skip: namespace traditionally has a dict pointing to self.
    pass

namespace = Namespace


# Component Event Model.
# todo: rename to `Instrument'
# todo: make Singleton behavior part of new, AutoInstrument class.
class Component(Singleton, Concretion, metaclass=Singleton.Meta):
    # Todo: Rename to Instrument?
    '''
    class Me(Component):
        Events = EVENTS + ['movementXYZ']

        class Module:
            # Or,
            implements = ['presidingAction']

            '''

    Module = 'bridge'
    Events = None

    def __init__(self, name, Module = None, Events = None, **others):
        # Register this component class instance with bridge module.
        self.Target = others.get('Target', self)

        if Module is None:
            Module = self.Module
        if Events is None:
            Events = self.getEventNames(Module)

        self.bindToRuntime(Module, Events)

        try: init = self.__instance_init__
        except AttributeError: pass
        else: init() # others['__init_args__']

    def __call__(self, ctlr, *args, **kwd):
        method = self.getTriggerFunction(ctlr.event_name)
        if callable(method):
            return method(ctlr, *args, **kwd)

    def __eq__(self, other):
        if self.sameClass(other):
            try:
                return self.__class__.__module__ == \
                         other.__class__.__module__ and \
                       self.__class__.__name__ == \
                         other.__class__.__name__

            except AttributeError:
                return False

    def getTriggerFunction(self, event_name):
        return getattr(self.Target, self.getTriggerName(event_name), None)
    def getTriggerName(self, event_name):
        return getTriggerName(event_name)

    def getEventNames(self, module):
        if self.Events is None:
            from ..events import getEventNames
            return getEventNames(module)

        return self.Events


    DYNAMIC_TRIGGERS = False
    def bindToRuntime(self, Module = None, Events = None):
        # system-dependent
        # Module: ExtensionPoint

        from ..events import Binding

        if Module == 'bridge':
            from stuphos import getBridgeModule
            Module = getBridgeModule()

        binding = Binding(Module)
        self.boundControllers = []

        if Events:
            for event in Events:
                # What this is saying is that it won't bind to events that don't exist.
                if self.DYNAMIC_TRIGGERS or callable(self.getTriggerFunction(event)):
                    ctlr = binding.getController(event)
                    ctlr.registerHandler(self)
                    self.boundControllers.append(ctlr)

    def __repr__(self):
        return '%s.%s (Component)' % (self.__module__, self.__class__.__name__)

    def __registry_delete__(self):
        for ctlr in self.boundControllers:
            ctlr.unregisterHandler(self)


def getTriggerName(event_name):
    return 'on%s%s' % (event_name[0].upper(), event_name[1:])

def newComponent(cls, name = None, **values):
    if name is None:
        # Unfortunately, it ends up taking up the module name that calls newClassObject.
        name = '%s.%s' % (cls.__module__, cls.__name__)

    # values['__instance_init__'] = cls.__init__
    # values['__init_args__'] = (args, kwd)
    return newClassObject(name, (Component, cls), values)

# This should be in events, but it relies on Singleton.
class DeclareEvent(Singleton, metaclass=Singleton.Meta):
    def __new__(self, *args, **kwd):
        return declareEventController(*args, **kwd)


# Memory management.
class writable(object):
    # Base class for emulated object attribute sets.
    def _setAttribute(self, name, value):
        return object.__setattr__(self, name, value)

class representable:
    def __repr__(self):
        return '<%s>' % self.__class__.__name__

class writeprotected(writable, representable):
    # The preferred native object base class for constraining memory sets.
    # The reason is derives from representable is so that you don't have to
    # derive from both of them.

    def __setattr__(self, name, value):
        # Note: allowing all private sets (preceeding undescore) simplifies Pythonic
        # private variables becoming writable.  Now this becomes unsuitable for any-
        # thing other than expressing objects into the virtual environment (where private
        # members are already protected).

        # try: pub = getattr(self, "__public_members__", "nothing")
        # except Exception as e:
        #     pub = f'{e.__class__.__name__}: {e}'
        # else:
        #     if pub == 'nothing' and name == '_owner': # Our case.
        #         debugOn()

        # print(f'{name} in {pub}')

        # So in otherwords, to repeat, I'm allowing private sets HERE, but only
        # because store_member denies them.
        if name.startswith('_') or name in getattr(self, '__public_members__', []):
            return self._setAttribute(name, value)

        raise AttributeError('Unwritable: %s' % name)

blockSetAttr = writeprotected

def getMemberStored(object, name):
    try: return object.__dict__[name]
    except KeyError:
        raise AttributeError(name)

class baseInstance(writable):
    def __setattr__(self, name, object):
        # from world import heartbeat as vm
        # task = vm.contextObject.task
        from stuphos.kernel import vmCurrentTask
        task = vmCurrentTask()

        try: existing = getMemberStored(self, name)
        except AttributeError:
            task.addObject(name)
        else:
            task.removeObject(existing)

        task.addObject(object)
        return self._setAttribute(name, object)


class extension:
    def __init__(self, object):
        self._object = object

wrapper = extension

class weakExtension(extension):
    def __init__(self, object = None, weakref = None):
        if weakref is None:
            weakref = ref.weakref(object)

        self._weakref = weakref

    @property
    def _object(self):
        return self._weakref()


def reprWrapper(self):
    return '<%s: type %s>' % (self.__class__.__name__,
                              self._object.__class__.__name__)


class extensionObject(writeprotected, Object, wrapper):
    # More opaque than a transparent object.
    __init__ = extension.__init__
    __repr__ = reprWrapper

    def _getExtensionOrObjectAttr(self, name):
        '''
        class remoteLibraryNativeExtension(extensionObject):
            """
            tool.remote-library:
                path: com/runphase/library-extension
                methods:
                    - object: stuphos.runtime.architecture.remoteLibraryNativeExtension._Install
                      name: remote

            interface:
                commands:
                    remoteInstall($method)::
                        def call(routine, args):
                            'com/runphase/library-extension/remote'()
                            return act(routine, args)

                        task = task$(call, $.components.object, args$())
                        return task.result()

                        """

            # Optionally:
            __getattr__ = extensionObject._getExtensionOrObjectAttr

            @classmethod
            def _Install(self, frame):
                # Do this once.
                task = frame.task
                task.native = self(task.native)

            def create(self, frame, path, *args):
                return log.library.create(path, *args) # *map(convertValue, args)

                '''

        try: return object.__getattribute__(self, name)
        except AttributeError:
            return getattr(self._object, name)


def recastObject(wrapper, value):
    # isinstance is unacceptable because wrapped objects might very well inherit from them:
    # Todo: omit list and dict?
    if type(value) in (str, bytes, list, tuple, dict, int, float):
        return value

    return wrapper._recast(value)

class wrappedObject(extensionObject):
    # Fully-wrapped.
    def __init__(self, defn, object):
        self._defn = defn
        extensionObject.__init__(self, object)

    @classmethod
    def _deny(self, typeClass, *names, **kwd):
        defn = kwd.get('defn', None)
        if defn is None:
            defn = dict()

        perClass = defn.setdefault(f'{typeClass.__module__}.{typeClass.__name__}', dict())
        for n in names:
            perClass[n] = False

        return defn

    def __getattr__(self, name):
        try: return object.__getattribute__(self, name)
        except AttributeError:
            if name == '_defn': # Unpickled? XXX Todo: figure out why.
                raise AttributeError(name)

            if not self._checkAccess(name):
                from stuphos.kernel import vmCurrentTask
                progr = vmCurrentTask().findProgrammer()
                raise NoAccessException('' if progr is None else progr.principal,
                                        resource = self._objectName().split('.') + [name])

            o = getattr(self._object, name)
            if callable(o):
                return self._wrappedFunction(self, o)

            return o

    def _objectName(self):
        t = type(self._object)
        return f'{t.__module__}.{t.__name__}'

    def _checkAccess(self, name):
        # print(f'{name}')
        try: access = self._defn[self._objectName()][name]
        except KeyError:
            return True

        return bool(access)

    def _recast(self, *args, **kwd):
        # Todo: return wrappingObject derivative based on value type
        # mapped to wrapper (self) property.
        return self.__class__(self._defn, *args, **kwd)


    class _wrappedFunction(writeprotected):
        def __init__(self, wrapper, object):
            self._wrapper = wrapper
            self._object = object

        def __call__(self, *args, **kwd):
            r = self._object(*list(self._unwrapArgs(args)),
                             **dict(self._unwrapKwd(kwd)))

            return recastObject(self._wrapper, r)

        def _unwrapArgs(self, args):
            for a in args:
                if isinstance(a, wrappedObject):
                    yield a._object
                else:
                    yield a

        def _unwrapKwd(self, kwd):
            for (name, value) in kwd.items():
                if isinstance(value, wrappedObject):
                    yield (name, value._object)
                else:
                    yield (name, value)


    def _call(self, proc, *args, **kwd):
        return self._recast(proc(*args, **kwd))


    def __call__(self, *args, **kwd):
        # Translate arguments.
        return self._wrappedFunction(self, self._object) \
            (*args, **kwd)

    def __dir__(self):
        # Todo: return self._call(dir, self._object)
        return dir(self._object)

    # __new__(cls[, ...])
    # __init__(self[, ...])
    # __del__(self)

    # __repr__(self)
    # __str__(self)

    # __bytes__(self)
    # __format__(self, format_spec)

    def __lt__(self, *args, **kwd):
        return self._object.__lt__(*args, **kwd)
    def __le__(self, *args, **kwd):
        return self._object.__le__(*args, **kwd)
    def __eq__(self, *args, **kwd):
        return self._object.__eq__(*args, **kwd)
    def __ne__(self, *args, **kwd):
        return self._object.__ne__(*args, **kwd)
    def __gt__(self, *args, **kwd):
        return self._object.__gt__(*args, **kwd)
    def __ge__(self, *args, **kwd):
        return self._object.__ge__(*args, **kwd)
    def __hash__(self):
        return self._object.__hash__()
    def __bool__(self):
        return self._object.__bool__()

    # __getattr__(self, name)
    # __getattribute__(self, name)
    # __setattr__(self, name, value)
    # __delattr__(self, name)
    # __dir__(self)
    # __get__(self, instance, owner=None)
    # __set__(self, instance, value)
    # __delete__(self, instance)
    # __slots__
    # __set_name__(self, owner, name)
    # __call__(self[, args...])
    def __len__(self):
        return self._object.__len__()
    def __length_hint__(self):
        return self._object.__length_hint__()
    def __getitem__(self, *args, **kwd): # key):
        return self._object.__getitem__(*args, **kwd)
    def __setitem__(self, *args, **kwd): # key, value):
        return self._object.__setitem__(*args, **kwd)
    def __delitem__(self, *args, **kwd): # key):
        return self._object.__delitem__(*args, **kwd)
    # __missing__(self, key)
    def __iter__(self):
        return self._object.__iter__()
    # __reversed__(self)
    def __contains__(self, *args, **kwd):
        return self._object.__contains__(*args, **kwd)
    def __add__(self, *args, **kwd):
        return self._object.__add__(*args, **kwd)
    def __sub__(self, *args, **kwd):
        return self._object.__sub__(*args, **kwd)
    def __mul__(self, *args, **kwd):
        return self._object.__mul__(*args, **kwd)
    def __matmul__(self, *args, **kwd):
        return self._object.__matmul__(*args, **kwd)
    def __truediv__(self, *args, **kwd):
        return self._object.__truediv__(*args, **kwd)
    def __floordiv__(self, *args, **kwd):
        return self._object.__floordiv__(*args, **kwd)
    def __mod__(self, *args, **kwd):
        return self._object.__mod__(*args, **kwd)
    def __divmod__(self, *args, **kwd):
        return self._object.__divmod__(*args, **kwd)
    # def __pow__(self, other[, modulo])
    def __lshift__(self, *args, **kwd):
        return self._object.__lshift__(*args, **kwd)
    def __rshift__(self, *args, **kwd):
        return self._object.__rshift__(*args, **kwd)
    def __and__(self, *args, **kwd):
        return self._object.__and__(*args, **kwd)
    def __xor__(self, *args, **kwd):
        return self._object.__xor__(*args, **kwd)
    def __or__(self, *args, **kwd):
        return self._object.__or__(*args, **kwd)
    def __radd__(self, *args, **kwd):
        return self._object.__radd__(*args, **kwd)
    def __rsub__(self, *args, **kwd):
        return self._object.__rsub__(*args, **kwd)
    def __rmul__(self, *args, **kwd):
        return self._object.__rmul__(*args, **kwd)
    def __rmatmul__(self, *args, **kwd):
        return self._object.__rmatmul__(*args, **kwd)
    def __rtruediv__(self, *args, **kwd):
        return self._object.__rtruediv__(*args, **kwd)
    def __rfloordiv__(self, *args, **kwd):
        return self._object.__rfloordiv__(*args, **kwd)
    def __rmod__(self, *args, **kwd):
        return self._object.__rmod__(*args, **kwd)
    def __rdivmod__(self, *args, **kwd):
        return self._object.__rdivmod__(*args, **kwd)
    # __rpow__(self, other[, modulo])
    def __rlshift__(self, *args, **kwd):
        return self._object.__rlshift__(*args, **kwd)
    def __rrshift__(self, *args, **kwd):
        return self._object.__rrshift__(*args, **kwd)
    def __rand__(self, *args, **kwd):
        return self._object.__rand__(*args, **kwd)
    def __rxor__(self, *args, **kwd):
        return self._object.__rxor__(*args, **kwd)
    def __ror__(self, *args, **kwd):
        return self._object.__or__(*args, **kwd)
    def __iadd__(self, *args, **kwd):
        return self._object.__iadd__(*args, **kwd)
    def __isub__(self, *args, **kwd):
        return self._object.__isub__(*args, **kwd)
    def __imul__(self, *args, **kwd):
        return self._object.__imul__(*args, **kwd)
    def __imatmul__(self, *args, **kwd):
        return self._object.__imatmul__(*args, **kwd)
    def __itruediv__(self, *args, **kwd):
        return self._object.__itruediv__(*args, **kwd)
    def __ifloordiv__(self, *args, **kwd):
        return self._object.__ifloordiv__(*args, **kwd)
    def __imod__(self, *args, **kwd):
        return self._object.__imod__(*args, **kwd)
    # __ipow__(self, other[, modulo])
    def __ilshift__(self, *args, **kwd):
        return self._object.__ilshift__(*args, **kwd)
    def __irshift__(self, *args, **kwd):
        return self._object.__irshift__(*args, **kwd)
    def __iand__(self, *args, **kwd):
        return self._object.__iand__(*args, **kwd)
    def __ixor__(self, *args, **kwd):
        return self._object.__ixor__(*args, **kwd)
    def __ior__(self, *args, **kwd):
        return self._object.__ior__(*args, **kwd)
    def __neg__(self):
        return self._object.__neg__()
    def __pos__(self):
        return self._object.__pos__()
    def __abs__(self):
        return self._object.__abs__()
    def __invert__(self):
        return self._object.__invert__()
    def __complex__(self):
        return self._object.__complex__()
    def __int__(self):
        return self._object.__int__()
    def __float__(self):
        return self._object.__float__()
    def __index__(self):
        return self._object.__index__()
    # __round__(self[, ndigits]):
    def __trunc__(self):
        return self._object.__trunc__()
    def __floor__(self):
        return self._object.__floor__()
    def __ceil__(self):
        return self._object.__ceil__()
    def __enter__(self):
        return self._object.__enter__()
    def __exit__(self, *args, **kwd): # exc_type, exc_value, traceback):
        return self._object.__exit__(*args, **kwd)
    # __match_args__
    def __await__(self):
        return self._object.__await__()
    def __aiter__(self):
        return self._object.__aiter__()
    def __anext__(self):
        return self._object.__anext__()
    def __aenter__(self):
        return self._object.__aenter__()
    def __aexit__(self, *args, **kwd): # exc_type, exc_value, traceback):
        return self._object.__aexit__(*args, **kwd)


wrappingObject = wrappedObject

class zonedObject(wrappedObject):
    def __init__(self, defn, object, taskId):
        self._taskId = taskId
        wrappedObject.__init__(self, defn, object)

    def _recast(self, object):
        return self.__class__(self._defn, object, self._taskId)

    def _checkAccess(self, name):
        if wrappedObject._checkAccess(self, name):
            from stuphos.kernel import vmCurrentTask
            return vmCurrentTask().id == self._taskId


# class NamespaceRedirect(extensionObject):
#     def __getattr__(self, name):
#         try: return object.__getattribute__(self, name)
#         except AttributeError:
#             try: return self._object[name]
#             except KeyError:
#                 if name == 'items':
#                     def items():
#                         o = self._object
#                         for name in dir(o):
#                             if not name.startswith('_'):
#                                 yield (name, getattr(o, name))

#                     return items # XXX :skip: Todo: safe_native

#                 raise AttributeError(name)


class Transparent(extensionObject):
    # Mainly, pass wrapped object value in resolveSymbol.
    pass

class Translucent(extensionObject):
    # Do not dereference ._object when passing/returning.
    pass

_transparent = Transparent
_translucent = Translucent

class ExceptionType(Translucent):
    # @classmethod
    def _fromMessage(self, message):
        return self._object(message)

class baseExceptionValue(Translucent):
    _type = None

    def __init__(self, value, type = None):
        Translucent.__init__(self, value)
        if type is not None:
            self._type = type

    def __repr__(self):
        if isinstance(self._object, type):
            return f'<{self.__class__.__name__}: type {self._object.__name__}>'

        return f'<{self.__class__.__name__}: type {self._object.__class__.__name__}: {self._object}>'
    vars()['repr$'] = property(__repr__)

    def __str__(self):
        return str(self._object)

    @property
    def typeName(self):
        return str(self._object.__class__.__name__)
    @property
    def typeMessage(self):
        return f'{self._object.__class__.__name__}: {self._object}'

    def __getattr__(self, name):
        try: return object.__getattribute__(self, name)
        except AttributeError as e:
            t = self._type
            if t is not None:
                if name in getattr(t, '_methods', []):
                    m = getattr(t, name)
                    return lambda *args, **kwd: \
                        m(self._object, *args, **kwd)

            # print(f'[evalue$attr] {t}: {self}: {name}')
            # debugOn()

            raise e


class safeNative(writeprotected):
    # XXX Not pickle-friendly.
    _nokwd = False

    def __init__(self, callable, name = None, nokwd = None):
        # self.__call__ = callable
        self._callable = callable
        self._name = name

        if nokwd is not None:
            self._nokwd = nokwd

    def __call__(self, *args, **kwd):
        # Note: this is done in Native._run
        # if self._nokwd:
        #     return self.__dict__['_callable'](*args)

        return self.__dict__['_callable'](*args, **kwd)

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(f'Attribute {repr(name)} must not started with underscore!')

        return getattr(self.__dict__['_callable'], name)

    # I don't know why I resort to this translation.  What am I hiding?
    # def __getattr__(self, name):
    #     # debugOn()
    #     try: return self.__dict__['_callable'][name] # translation
    #     except (TypeError, KeyError):
    #         raise AttributeError(name)

    def __getitem__(self, item):
        return self.__dict__['_callable'][item]

    def __repr__(self):
        return f'<native {self.__name__}>' # repr(self._callable)

    @property
    def __name__(self):
        # print(f"safe-native: {self.__dict__['_callable']}")
        if self._name is not None:
            return self._name

        return self.__dict__['_callable'].__name__

    @property
    def __doc__(self):
        # print(f"safe-native: {self.__dict__['_callable']}")
        return self.__dict__['_callable'].__doc__

_safe = _safe_native = safeNative


class safeNativeNoKeyword(safeNative):
    _nokwd = True

_safe_native_nokwd = safeNativeNoKeyword


class safeNativeClass(object):
    def __getattribute__(self, name):
        o = object.__getattribute__(self, name)
        if name.startswith('_'):
            return o

        if callable(o):
            return _safe_native(o)

        return o

_safe_native_class = safeNativeClass

class safeNativeObject:
    def __init__(self, object):
        self._object = object

    # getattribute is required, but it catches .__dict__ (and probably .__class__)
    def __getattribute__(self, name):
        # print(f'safe-native.getattr: {name}')
        # debugOn()

        if name in ['__dict__', '__class__', '_object']:
            return object.__getattribute__(self, name)

        o = getattr(object.__getattribute__(self, '__dict__')['_object'], name)
        if callable(o):
            return _safe_native(o)

        return o

    def __repr__(self):
        try: name = self._object.__name__
        except AttributeError:
            name = self._object.__class__.__name__

        return f'<native-object {name}>'

    def __iter__(self):
        return self._object.__iter__()
    def __next__(self):
        return self._object.__next__()

_safe_native_object = safeNativeObject


class safeIterator(writeprotected):
    __init__ = safeNativeObject.__init__
    __repr__ = safeNativeObject.__repr__

    def __iter__(self):
        return iter(self._object)

    def __next__(self, *args, **kwd):
        return self._object.__next__(*args, **kwd)

    @classmethod
    def decorator(self, function):
        def call(*args, **kwd):
            return self(function(*args, **kwd))
        return call

_safe_iterator = safeIterator


# Execution Primitives.
class InterpreterState(Exception):
    pass

class Yield(InterpreterState): # virtual.Machine.Task.Yield?
    pass

class Continuation(Yield):
    # Used to signal that processing should just continue.
    pass

class AwaitableContinuation(Continuation):
    pass

class OuterFrame(Continuation):
    def __init__(self, frame = None):
        self.frame = frame

    @property
    def returnValue(self):
        return self.frame.returnValue

    def onComplete(self, *args, **kwd):
        return self.frame.onComplete(*args, **kwd)

class BypassReturn(Continuation):
    pass

class YieldFrame(OuterFrame):
    pass


class Interrupt(InterpreterState):
    _uncatchable = True
class Timeout(Interrupt):
    _uncatchable = False
class TimeoutQuotaExceeded(Timeout):
    _uncatchable = True


class Uncompleted(RuntimeError):
    "A native or didn't complete because it resulted in an asynchronous operation that wasn't handled."


class Interface(writeprotected, Object):
    # Revealed to the GIRL object runtime, proxies the local subroutine def
    # through the vm invocation instruction.
    class _Meta(Object._Meta):
        Attributes = [('procedure', lambda i: repr(i._procedure))]

    def __init__(self, procedure):
        self._procedure = procedure

    def __str__(self):
        return self._procedure.pathString()

    @property
    def _activation(self):
        # Return a copy of a PC-tracking invocation object.
        # Todo: Invocation(self._procedure)
        return self._procedure

class Done(InterpreterState, Object):
    pass

class ScheduledProcedureMixin:
    def _runSchedule(self, active, frame):
        return self._run(frame)

class Procedure(Object, ScheduledProcedureMixin):
    _Interface = Interface
    _Done = Done

    def __resolve__(self):
        return repr(self)


class Computer:
    # Instruction set.
    pass

class Source:
    pass


class Generator(writeprotected):
    _position = None

    def __init__(self, program, frame):
        self._setAttribute('_program', program)
        self._setAttribute('_frame', frame) # Old frame.
        self._setAttribute('_task', frame.task)
        self._setAttribute('_stack', dict())

    def __repr__(self):
        return f'<generator {repr(self._program)}>'

    def isGenerator(self):
        return True

    def __iter__(self):
        return self

    def __stopIteration__(self, pos):
        self._setAttribute('_position', pos)

    def _deleteStack(self, task):
        taskId = task.id
        stack = self._stack[taskId]
        del self._stack[taskId]

        ro = task.removeObject
        for i in stack:
            ro(i)

    def _restoreStack(self):
        stack = self._task.stack
        taskId = self._task.id

        try: saved = self._stack[taskId]
        except KeyError: pass
        else:
            list.extend(stack, saved)
            self._stack[taskId]

    def _saveStackRemove(self, frame):
        stack = self._task.stack
        n = len(stack) - frame.initialStackSize

        # Since this function is called by levelStack when the frame
        # is remove, yield will already have put a value on the stack,
        # which needs to not be saved (on this generator) or restored.
        x = self._stack[self._task.id] = stack[-n:-1]

        # print(f'[gen.stack.save] {x}')
        # debugOn()

        list.__delitem__(stack, slice(-n, -(n-1)))

    def __next__(self, genCycle = None, knowsYield = False):
        # Since for_next calls this function, we set up a new
        # frame call and then raise a Yield so that the runtime
        # can rely on frame-stack state to push the next iter.
        #
        # Also, memory.native.sequence calls this function
        # (setting genCycle).

        if not knowsYield:
            raise AssertionError('Must call Generator.__next__(knowsYield = True)!')

        self._restoreStack()

        # todo: do general frame copy, storing parameters on instance construction.
        new = self._task.addFrameCall \
            (self._program,
             locals = self._frame.locals,
             environ = self._frame.environ,
             programmer = self._frame.programmer,
             arguments = self._frame.arguments,
             keywords = self._frame.keywords,
             genContinue = self)

        # print(f'[gen.next] {new}')

        @new.onComplete
        def completion(it, etype = None, value = None, tb = None, traceback = None):
            # print(f'[gen.next.complete] {value}')

            it.genContinue = False # so that levelStack doesn't restore.

            if etype is StopIteration:
                if callable(genCycle):
                    # The call path for natives like sequence and kernel/map.
                    genCycle(self, stopIteration = True)
                else:
                    self._task.onStopIteration(it, self._position)

                return True

            # Note: don't pass/raise exceptions other than StopIteration.
            elif value is None and callable(genCycle):
                genCycle(self)

        raise new.yieldOuter


def progr_reveal(progr):
    hide = getattr(progr, '_principalHide', False)
    if hide:
        return f'<{"hidden" if hide is True else hide}>'

    try: p = progr.principal # WithCase?
    except AttributeError:
        p = progr

    p = str(p)

    if p.startswith('apikey:'):
        return '<API-Key>'

    return p

class NoAccessException(Exception):
    @classmethod
    def _fromMessage(self, args):
        (path, access) = args
        return self(None, resource = path, access = access)

    def __init__(self, programmer, resource = '?', access = '?', task = None):
        # todo: use task.id?
        # This format is compatible with 'agent grant' command.
        path = resource if isinstance(resource, str) else ' '.join(map(str, resource))
        Exception.__init__(self,
            f"[{task.taskName if task else ''}] {progr_reveal(programmer)} {access} {path}")
        # '[%s] %s for %s by %r' % (task.taskName if task else '', access, resource, programmer)

        self.programmer = programmer
        self.resource = resource
        self.access = access
        self._task = task

    # def pathNotFound(self):
    #     raise PathNotFound(self.resource)


# Todo: because PathNotFound is part of ph, move this into that package.
def pathNotFound(noAccess):
    raise ValueError(noAccess.resource) # PathNotFound(noAccess.resource)

NoAccessException.pathNotFound = pathNotFound

def recastNoAccess(function):
    '''
    Catch NoAccessExceptions and reraise them as PathNotFound.
    This hides the fact that a missing permission can't be used
    to verify an existing node in the library.

    Use as a decorator.

    '''

    def call(*args, **kwd):
        try: return function(*args, **kwd)
        except NoAccessException as e:
            pathNotFound(e)

    call.__name__ = 'recastNoAccess:' + function.__name__
    return call


# Permissions.
SPECIAL_PRINCIPALS = ['group', 'agent', 'role', 'apikey']

def isSpecialPrincipal(principal):
    i = principal.find(':')
    if i > 0:
        return principal[:i].lower() in SPECIAL_PRINCIPALS


# Game-Level Objects.
class UnknownFlag(NameError):
    pass

class Bitvector(Object):
    # A pure implementation of the bitvector type in game module.
    class _Meta(Object._Meta):
        Attributes = Object._Meta.Attributes + ['set']

    def __init_subclass__(self, **kwd):
        pass

    def __init__(self, __bitv = 0, **bits):
        # This is an abstract base class.
        assert self.__class__ is not Bitvector

        self.__bitv = int(__bitv)
        for (name, value) in bits.items():
            setattr(self, name, bool(value))

        self.getUpperBitvNames()

    @classmethod
    def getUpperBitvNames(self):
        try: return self.__UPPER_BITVECTOR_NAMES
        except AttributeError:
            names = self.__UPPER_BITVECTOR_NAMES = \
                [n.upper() for n in self.BITVECTOR_NAMES]

        return names

    BITVECTOR_NAMES = []

    def isBitSet(self, bit):
        return bool(self.__bitv & bit)
    def getFlagBit(self, name):
        try: return (1 << self.getUpperBitvNames().index(name.upper()))
        except ValueError:
            raise UnknownFlag

    def isFlagSet(self, name):
        return self.isBitSet(self.getFlagBit(name))

    @property
    def names(self):
        return self.BITVECTOR_NAMES

    @property
    def set(self):
        return [name for name in self.names if self.isFlagSet(name)]

    @property
    def notset(self):
        return [name for name in self.names if not self.isFlagSet(name)]

    unset = nonset = notset

    def __getattr__(self, name):
        try: return self.isFlagSet(name)
        except UnknownFlag:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        try: bit = self.getFlagBit(name)
        except UnknownFlag: return object.__setattr__(self, name, value)
        else: self.__bitv |= bit if value else ~bit

    def __int__(self):
        return int(self.__bitv)
    def __str__(self):
        return ', '.join(map(str, self.set))
    def __iter__(self):
        return iter(self.set)

class PromptPreferences(Bitvector):
    BITVECTOR_NAMES = ['Mail', 'DataRate']



# FTP Module.
from stuphmud.server.player import playerAlert, getCapitalizedName
from stuphmud.server.player.db.managedfs import getCheckedPlayerFilename, PERSONAL_PLRFILE_DIR, KEYFILE_DIR

from os import rename, unlink, stat, listdir
from os.path import dirname

from errno import EEXIST

class FTPDispatchModule:
    @classmethod
    def getDispatchAction(self, operation):
        actionName = '_do%s' % getCapitalizedName(operation)
        if hasattr(self, actionName):
            return self(actionName)

    def __init__(self, operation):
        self.operation = operation
        self.invoke = getattr(self, operation, None)

    def __call__(self, peer, *params):
        # Do operation.
        if params:
            filename = params[0]
            params = params[1:]
        else:
            filename = ''

        plrname = peer.avatar and peer.avatar.name.lower() or ''
        dirOp = bool(getattr(self.invoke, 'dirOperation', False))
        plrfile = getCheckedPlayerFilename(peer, filename, PERSONAL_PLRFILE_DIR, dirOp)
        self.invoke(peer, plrname, plrfile, *params)

        # Todo: playerAlert this?
        print('Operation [%s] %r complete!' % (self.operation, plrfile), file=peer)

    # Operations.
    def _doCreate(self, peer, plrname, plrfile, content):
        # Todo: expect checksum
        EnsureDirectories(plrfile)
        fl = open(plrfile, 'w')
        fl.write(content)
        fl.flush()
        fl.close()

    def _doAppend(self, peer, plrname, plrfile, content):
        fl = open(plrfile, 'a')
        fl.write(content)
        fl.flush()
        fl.close()

    def _doDelete(self, peer, plrname, plrfile):
        unlink(plrfile)

    def _doRename(self, peer, plrname, plrfile, newname):
        newname = getSecureFilename(newname)
        if not newname:
            playerAlert('Unable to produce secure filename.')

        newname = getPlayerFilename(plrname, newname)
        if not newname:
            playerAlert('Could not obtain player filename.')

        rename(plrfile, newname)

    def _doTouch(self, peer, plrname, plrfile):
        playerAlert('Touch operation not yet supported.')

    def _doTruncate(self, peer, plrname, plrfile):
        fl = open(plrfile, 'w')
        fl.truncate(0)
        fl.flush()
        fl.close()

    def _doList(self, peer, plrname, plrfile = None):
        from stuphos.etc.tools import columnize
        try: files = listdir(plrfile)
        except OSError as e: playerAlert(str(e))
        else: playerAlert(columnize(files, 2, 40))
    _doList.dirOperation = True

    def _doStat(self, peer, plrname, plrfile):
        st = stat(plrfile)
        print('%s: %d' % (plrfile, st.st_size), file=peer)

    def _doView(self, peer, plrname, plrfile):
        peer.page_string(open(plrfile).read())

    def _doLoad(self, peer, plrname, plrfile):
        from pickle import dump
        from io import StringIO
        from pickletools import dis

        from stuphos.language.document import interface as document

        env = document.Environment()
        loader = document.Loader(env)
        loader.loadDocument(document.FileSource(plrfile))

        buf = StringIO()
        buf2 = StringIO()

        dump(env.olc, buf)
        buf.seek(0)

        dis(buf, buf2)
        peer.page_string(buf2.getvalue())

    def _doRender(self, peer, plrname, plrfile):
        # as django template...
        pass

    def _doDecrypt(self, peer, plrname, plrfile):
        # Todo: multithread this! (Decryption takes time)
        keyfile = getCheckedPlayerFilename(peer, 'my.key', KEYFILE_DIR)
        key = LoadKeyfile(keyfile) # Catch IOError and report.
        DecryptFile(key, plrfile, armored = True)
        playerAlert('Decrypted [%s]' % plrfile)

def EnsureDirectories(filename):
    # This creates the directories leading up to the base.
    from os.path import split as splitpath
    (path, fn) = splitpath(filename)
    if not fn:
        raise ValueError('Directory path has no base filename! (%s)' % filename)

    from os import makedirs
    try: makedirs(path)
    except OSError as e:
        if e.args[0] != EEXIST:
            raise

def LoadKeyfile(filename):
    # Format: plain ascii keyphrase...
    return open(filename).read().strip()

def getCipher(key):
    from crypto.cipher.trolldoll import Trolldoll
    alg = Trolldoll(ivSize = 160)
    alg.setPassphrase(key) # .setKey(key)
    return alg

def encrypt(key, content):
    return getCipher(key).encrypt(content)
def decrypt(key, content):
    return getCipher(key).decrypt(content)

def encryptArmored(key, content):
    return encrypt(key, content).encode('base64').replace('\n', '')
def decryptArmored(key, content):
    return decrypt(key, content.decode('base64'))

def DecryptFile(key, filename, armored = False):
    # Decrypt file in place.
    content = open(filename).read()
    decrypted = decryptArmored(key, content) if armored else decrypt(key, content)

    # Todo: decrypt to migrating file in blocks..
    tmp = filename + '.migrating'
    fl = open(tmp, 'wt')
    fl.write(decrypted)
    fl.flush()
    fl.close()
    rename(tmp, filename)



'Schema for a StuphMUD player save - point.'

from sqlobject import *

class PlayerStore(SQLObject):
	# This is introduced by playerSavePoint - Rename to savepoint.
	save_point_indication = StringCol()

	name            = StringCol() # limit this length
	idnum           = IntCol()
	level           = IntCol(default = 0)
	pwd             = StringCol() # limit this length
	host            = StringCol(default = '') # limit this length
	email           = StringCol(default = '') # limit this length

	color           = BLOBCol(default = '')
	talks           = BLOBCol(default = '')
	skills          = BLOBCol(default = '')
	affected        = BLOBCol(default = '')

	birth           = IntCol(default = 0)
	sex             = IntCol(default = 0)
	exp             = IntCol(default = 0)
	played          = IntCol(default = 0)

	loadroom        = IntCol(default = -1)
	saveroom        = IntCol(default = -1)

	invis           = IntCol(default = 0)
	freeze          = IntCol(default = 0)
	badpws          = IntCol(default = 0)

	lastlogin       = IntCol(default = 0)
	lastlogon       = IntCol(default = 0)

	cha             = IntCol(default = 11)
	con             = IntCol(default = 11)
	dex             = IntCol(default = 11)
	str             = IntCol(default = 11)
	wis             = IntCol(default = 11)
	intel           = IntCol(default = 11)
	stradd          = IntCol(default = 0)

	act             = IntCol(default = 0)
	disp            = IntCol(default = 0)
	pref            = IntCol(default = 0)
	pref2           = IntCol(default = 0)
	affectedby      = IntCol(default = 0)

	dts             = IntCol(default = 0)
	deaths          = IntCol(default = 0)
	mkills          = IntCol(default = 0)
	pkills          = IntCol(default = 0)

	maxhit          = IntCol(default = 10)
	hit             = IntCol(default = 10)
	maxmana         = IntCol(default = 10)
	mana            = IntCol(default = 10)
	move            = IntCol(default = 10)
	maxmove         = IntCol(default = 10)

	bank            = IntCol(default = 0)
	gold            = IntCol(default = 0)

	clan            = IntCol(default = -1)
	clanrank        = IntCol(default = 0)

	race            = IntCol(default = 0)
	chclass         = IntCol(default = 0)

	spouse          = IntCol(default = -1)
	maritalstatus   = IntCol(default = 0)
	num_marriages   = IntCol(default = 0)

	qpoints         = IntCol(default = 0)
	remort          = IntCol(default = 0)
	hometown        = IntCol(default = -1)
	remortpoints    = IntCol(default = 0)

	alignment       = IntCol(default = 0)
	height          = IntCol(default = 120)
	weight          = IntCol(default = 120)

	pagelength      = IntCol(default = 22)

	armor           = IntCol(default = 10)
	damroll         = IntCol(default = 0)
	hitroll         = IntCol(default = 0)
	wimp            = IntCol(default = 0)

	drunk           = IntCol(default = 0)
	hunger          = IntCol(default = 0)
	thirst          = IntCol(default = 0)

	savingthrow1    = IntCol(default = 0)
	savingthrow2    = IntCol(default = 0)
	savingthrow3    = IntCol(default = 0)
	savingthrow4    = IntCol(default = 0)
	savingthrow5    = IntCol(default = 0)

	spellattack1    = IntCol(default = -1)
	spellattack2    = IntCol(default = -1)

	spells2learn    = IntCol(default = 0)



# Core Content Design and Delivery Client
# Todo: move into stuphos.management
# XMLRPC Python API -- requires WRLC
from xmlrpc.client import ServerProxy, MultiCall
from urllib.parse import quote
import ssl
import sys

import op # actionable


HOSTNAME = 'localhost'
PORT = 2172

HOST = '%s:%s' % (HOSTNAME, PORT)
PROTOCOL = 'https'

# client XMLRPC+HTTPS: TLSv1 Unknown Certificate Authority
# [SSL: TLSV1_ALERT_UNKNOWN_CA] tlsv1 alert unknown ca (_ssl.c:590)
# [SSL: HTTP_REQUEST] http request (_ssl.c:590)
# SSLError: [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed (_ssl.c:590)

try: namespace
except NameError:
    namespace = dict

auth = login = namespace

class folderUpload(namespace):
    'Package up a directory structure for deployment upload.'

    @classmethod
    def build(self, path, exclude = []):
        if isinstance(exclude, str):
            import re
            m = re.compile(exclude).match

            def exclude(s):
                return m(s) is not None

        elif isinstance(exclude, (list, tuple)):
            exclude = contains(exclude)

        def node(path, relative, contentName, **kwd):
            content = path.platformMapped # .read()
            return dict(path = relative, name = path.basename,
                        **dict(kwd, **{contentName: content}))

        def structure(p, r):
            return node(p, r, 'document')
        def activities(p, r):
            return node(p, r, 'source')

        def document(p, r):
            typeContent = 'application/binary' # todo: get from mimeType[p.extension]
            return node(p, r, 'content', typeContent = typeContent)

        def c(path, type, relative):
            s = []
            a = []
            k = []
            d = []

            for p in path.listing:
                if p.isdir:
                    r = relative + p.basename

                    if p.basename == 'structures':
                        s.append(c(p, structure, r))
                    elif p.basename == 'activities':
                        a.append(c(p, activity, r))
                    else:
                        k.append(c(p, document, r))

                elif not exclude(p):
                    d.append(type(p, r))

            return self(structures = s, activities = a,
                        folders = k, documents = d)

        return c(path, document, '')

    @property
    def package(self):
        # todo: return a package suitable for upload.
        pass


class phClientFS:
    '''
    client.upload(phClientFS(args).structure,
                  path = 'mount/point')

    '''

    def __init__(self, path):
        self.path = io.path(path)

    @property
    def structure(self):
        if self.path.isdir:
            new = classOf(self)
            return dict((path.basename, new(path).structure)
                        for path in self.path.listing)

        return self.path.read()


def deploy(client, path, *args, **kwd):
    '''
    # python phClient -u me -p none deploy repository/deploy projectName

    def deploy(client, path, *args, **kwd):
        script = r'.*repository/deploy\.py$' # This shouldn't be necessary...
        client.root += client.folderUpload(path.folder(*args), script)

        '''

    path = io.here(path)
    e = path.extension

    if not e:
        path = path.pyOf
    else:
        assert e == 'py'

    try: d = path.loading.module.deploy
    except AttributeError:
        o = dict(client = client, path = path,
                 args = args, kwd = kwd)

        exec(compile(open(path, "rb").read(), path, 'exec'), globals(), o)
    else:
        return d(client, path, *args, **kwd)


class apiClient:
    @classmethod
    def Main(self, argv = None):
        # See phClient.__main__
        # if __name__ == '__main__': from phClient import apiClient; apiClient.Main()

        from optparse import OptionParser
        parser = OptionParser()
        parser.add_option('--host', default = HOST)
        parser.add_option('--protocol', default = PROTOCOL)
        parser.add_option('--username') # principal
        parser.add_option('--configuration', '-C')
        parser.add_option('--argument-file', '-a', default = list(), action = 'append')
        parser.add_option('--wget-bin')
        parser.add_option('--wget-arg', dest = 'wget_args', default = list(), action = 'append')
        parser.add_option('-c', '--remote-script')
        parser.add_option('--remote-script-path', default = 'script/execute')
        (options, args) = parser.parse_args(argv)

        def buildClient():
            if options.configuration:
                password = open(options.configuration).read().strip()
            else:
                from getpass import getpass
                password = getpass()

            return self(options.username, password,
                        host = options.host,
                        protocol = options.protocol)


        if args:
            import json

            cmd = args[0]
            args = args[1:]

            if cmd == 'call':
                if args:
                    # AUTH = "--username=op -C ~/.ph/runcore/op/password"
                    # SERVICE = "components/services/lisp/do"
                    # INPUT = "ph-runcore-command.lsp"
                    # python -m phClient.cli "$AUTH" call "$SERVICE" -a "$INPUT"

                    args = args[1:] + [open(f).read() for f in options.argument_file]
                    print(buildClient().call(*args))

            elif cmd == 'search':
                if args:
                    for (section, results) in buildClient().search(*args).items():
                        print('%s:' % section)
                        print(indent(json.dumps(results)))
                        print('')

            elif cmd == 'deploy':
                if args:
                    deploy(buildClient(), *args, options = options)

            elif cmd == 'upload':
                if args:
                    pass

            elif cmd == 'script':
                import op

                script = options.remote_script
                if script:
                    '''
                    # Browse 'public' permissions in json format.
                    echo public | python3 -m stuphos.phClient script \
                        -c "return 'text/json/dumps'(permissions(), 1)" | less

                    '''

                    assert not args
                    script += '\n'

                else:
                    '''
                    python3 -m stuphos.phClient script -C ~/.thetaplane.apikey \
                            --host thetaplane.com -- << EOF

                        return 'text/json/dumps'(permissions())

                    EOF
                    '''

                    (script,) = args

                    if script == '--':
                        from sys import stdin
                        script = stdin.read()
                    else:
                        script = open(script).read()


                if options.configuration:
                    key = open(options.configuration).read()
                else:
                    from getpass import getpass
                    key = getpass('API Key: ')


                wget = io.path(options.wget_bin or '/usr/bin/wget')

                url = f'{options.protocol}://{options.host}/{options.remote_script_path}'

                wgetArgs = (url, '--post-data', f'key={quote(key)}&script={quote(script)}',
                            '-O', '-', '-q', '--content-on-error')

                wgetArgs += tuple(options.wget_args)

                try: r = wget.pipe(*wgetArgs)
                except op.platform.path.CalledProcessError as e:
                    print(f'*** {e.__class__.__name__}', file = sys.stderr)
                    # print(f'  {" ".join(e.cmd)}')
                    print(e.stderrOutput.decode() + '\n' + e.stdOutput.decode(), file = sys.stderr)
                else:
                    print(r.decode())


    def __init__(self, username, password, host = HOST, protocol = 'https', runtime = None, **kwd):
        # Note: xmlrpclib._Method.__call__ does not accept keyword arguments.
        if username is None and password is None:
            endpoint = '%s://%s/rpc' % (protocol, host)
        else:
            endpoint = '%s://%s:%s@%s/rpc' % (protocol, username, password, host)

        self.host = host
        self.protocol = protocol

        self.endpoint = endpoint

        self.runtime = runtime if runtime is not None else globals()['runtime'](**kwd)

    @property
    def server(self):
        if self.protocol == 'https':
            return ServerProxy(self.endpoint, context = ssl.SSLContext())

        return ServerProxy(self.endpoint)

    def run(self, *args, **kwd):
        return self.runtime.run(self, *args, **kwd)

    def call(self, method, *args, **kwd):
        return self.server.library.callMethod(method, *args, **kwd)
    __call__ = call

    @property
    def multiCall(self):
        return MultiCall(self.server)

    def search(self, *queryArgs):
        return self.call('core/interface/kernel/librarySearch', *queryArgs)

    def evaluate(self, expression, *args, **kwd):
        return self.server.engine.evaluateAgent(expression, *args, **kwd)


    # Local Emulations.
    def moduleCompile(self, source):
        syntax = self.server.engine.moduleCompile(source)
        return self.syntaxObject(syntax, source = source)

    def moduleAssemble(self, source):
        return self.server.engine.moduleAssemble(source)

    def pathCompile(self, path):
        syntax = self.server.engine.pathCompile(path)
        return self.syntaxObject(syntax, path = path)

    def pathAssemble(self, path):
        return self.server.engine.pathAssemble(path)


    # Object Model.
    def _queryNodeData(self, *args, **kwd):
        return self.server.library.queryNode(*args, **kwd)

    def queryNode(self, *args, **kwd):
        return self.Node(self, self._queryNodeData(*args, **kwd))
    __getitem__ = query = queryNode

    root = property(queryNode)

    def loadChildrenMultiCall(self, children):
        data = self.multiCall
        q = data.library.queryNode

        for c in children:
            q(c.path)

        for (i, c) in data():
            r[i].load(c)

        return children

    def loadChildrenIndividually(self, children):
        for c in children:
            c.load(self._queryNodeData(c.path))

        return children

    loadChildren = loadChildrenIndividually # loadChildrenMultiCall

    def __iter__(self):
        return iter(self.root)

    def node(self, type, **data):
        data['type'] = type
        return self.Node.FromClientAndData(self, **data)

    def folder(self, **data):
        return self.node('directory', **data)
    def module(self, **data):
        return self.node('module', **data)
    def structure(self, **data):
        return self.node('structure', **data)
    def package(self, **data):
        return self.node('package', **data)

    def document(self, **data):
        try: data['content_type'] = data.pop('typeContent')
        except KeyError: pass
        return self.node('document', **data)

    def upload(self, structure, **data):
        return self.node('upload', structure = structure, **data)


    def folderUpload(self, *args, **kwd):
        return self.upload(folderUpload.build(*args, **kwd).package)


    class Node(namespace):
        '''
        client['my/components'] += client.folder \
            (name = 'program', children = \
             [client.structure(name = 'interface', document = '...'),
              client.module(name = 'object', program = '...')])

        '''
        # todo: figure out when exactly nameSlashed is used (read) and not path.

        @classmethod
        def FromClientAndData(self, client, **data):
            # This should be compatible with data from 'exported' node queries,
            # and also from the client.node() method.  But, if you have data
            # directly from exported, use the constructor and skip this method.
            try: path = data['path'] # XXX todo: path as dict item has no meaning... pop?
            except KeyError:
                try: path = data['nameSlashed']
                except KeyError:
                    # An often-used code point.  Note, this gets called for children
                    # of the root node, since no path is supplied.
                    return self(client, **data)
                else:
                    # This _should_ correspond with the trailing part of
                    # nameSlashed.  But if nameSlashed is set and path
                    # is not, then ignore name (which probably shouldn't be set).
                    #
                    # Again: nameSlashed probably shouldn't be used, either...
                    # because it confuses the namespace constructor: it should
                    # be popped from 'data', since it's being passed via keyword.

                    del data['name']

            segments = path
            if isinstance(segments, str):
                segments = segments.split('/')

            # Q: Why was I trying to calculate 'parent'?
            # Parent is derived from nameSlashed, which is set by path...

            try: name = data.pop('name')
            except KeyError:
                name = segments[-1]
                # parent = segments[:-1]

                # If there's no name, then either nameSlashed or path sets
                # the whole name, necessarily.
            else:
                # parent = segments

                # Note: since name is set, if nameSlashed is not set,
                # path must necessarily be set.  But, it is the parent
                # path, so, rebuild the symbol here as a whole name.

                path = '/'.join(segments + [name])

            return self(client, nameSlashed = path,
                        name = name, **data)

        def __init__(self, client, *args, **kwd):
            namespace.__init__(self, *args, **kwd)
            self.client = client

        def __repr__(self):
            # name = nameOf(classOf(self))
            try: return '%s: %s' % (self.type, self.path)
            except TypeError:
                # Unloaded.
                return 'unloaded: %s' % self.name

        def __str__(self):
            return self.name

        @property
        def path(self):
            try: return self.nameSlashed
            except AttributeError:
                # Assume child of root.
                return self.name

        @property
        def parent(self):
            path = self.nameSlashed.split('/')
            if len(path) < 2:
                # Child of root node.
                return '' # None # XXX Must be marshallable!

            return '/'.join(path[:-1])

        @parent.setter
        def parent(self, path):
            if 'nameSlashed' in self:
                raise AttributeError('parent already set; delete it first')

            # Upload nodes may have no names.  These kinds of nodes, when
            # added to their parents, may function relative to the parent.
            try: name = self['name']
            except KeyError: pass
            else:
                if path is not None:
                    path += '/'
                    path += name
                else:
                    path = name

            self.nameSlashed = path

        @parent.deleter
        def parent(self):
            del self.nameSlashed

        def load(self, data = None):
            # Q: Shouldn't this set loaded state upon success?
            try: del self['unloaded']
            except KeyError: pass
            else:
                if data is None:
                    data = self.client._queryNodeData(self.path)

                self.update(data)

            return self
        loaded = property(load)

        def create(self):
            if self.type == 'directory':
                assert self.client.server.library.addFolder(self.parent, self.name)
                self.createChildren()

            elif self.type == 'module':
                assert self.client.server.library.addModule(self.parent, self.name, self.program)
            elif self.type == 'structure':
                assert self.client.server.library.addStructure(self.parent, self.name, self.document)
            elif self.type == 'package':
                assert self.client.server.library.addPackage(self.parent, self.name, self.source)
            elif self.type == 'document':
                assert self.client.server.library.addDocument(self.parent, self.name, self.typeContent, self.content)
            elif self.type == 'upload':
                assert self.upload(self.structure)
            else:
                raise TypeError(self.type)

            return self

        def createChildren(self):
            for node in self.children:
                del node.parent
                self += node

        def add(self, node, name = None):
            if name is not None:
                node.name = name

            node.parent = self.path
            node.create() # XXX todo: if not exists?  Although, this wouldn't be called if it did...

        __add__ = add

        def __iadd__(self, node):
            self.add(node)
            return self

        def __getattr__(self, name):
            try: return object.__getattribute__(self, name)
            except AttributeError as e:
                if name in ('nameSlashed', 'path'):
                    raise e

                try: t = self['type']
                except KeyError:
                    raise TypeError('Untyped node')

                if t == 'module':
                    return self._call(self, name)
                if t == 'directory':
                    return self.getChild(name)

                raise e

        class _call(object):
            def __init__(self, node, name):
                self.node = node
                self.name = name

            def __call__(self, *args):
                return self.node.client.call(self.name, *args)

        def __iter__(self):
            if self.type == 'directory':
                return iter(self.preloadChildren())

            i = client.server.engine.evaluateAgent
            i = i('symbols(node)', node = self.path)
            return iter(i)

        @property
        def children(self):
            if self.type != 'directory':
                raise TypeError('Not a directory')

            return self.preloadChildren()

        def preloadChildren(self):
            try: e = self['children']
            except KeyError:
                return []

            node = classOf(self); client = self.client
            try: path = self.path
            except AttributeError:
                path = None

            for (n, c) in enumerate(e):
                if isinstance(c, str):
                    c = dict(name = c, unloaded = True)
                    if path is not None:
                        c['path'] = path

                    c = self.FromClientAndData(client, **c)
                    e[n] = c

            return e

        @property
        def childrenLoaded(self):
            # Actually retrieve data from server for each child...
            return self.client.loadChildren(self.children)


        @property
        def childrenNames(self):
            return (node.name for node in self.children)

        @property
        def symbols(self):
            return iter(self)

        def getChild(self, name):
            for node in self.children:
                if node.name == name:
                    return node

        __div__ = getChild
        # __getitem__ = getChild


        # Conventional hacks for document type:
        @property
        def typeContent(self):
            try: return self['typeContent']
            except KeyError:
                try: return self['content_type']
                except KeyError:
                    raise AttributeError('content_type/typeContent')

        @typeContent.setter
        def typeContent(self, value):
            self['content_type'] = value

        content_type = typeContent


        def packageUpload(self, structure): # XXX naming
            # note that 'package' in this case is a document structure.
            # todo: something like client.folder('path/to') += client.upload(dict(documents = dict(structures = dict(one = ''))))
            def internalizeType(o):
                if isinstance(o, dict):
                    return dict((n, internalizeType(v)) for (n, v) in o.items())
                elif isinstance(o, (list, tuple)):
                    return list(map(internalizeType, o))

                if not isinstance(o, (int, float, str)): # complex?
                    raise TypeError(o.__class__.__name__)

                return o

            return self.client.server.library.uploadStructure \
                    (self.path, internalizeType(structure))

        upload = packageUpload


    def login(self, playerName, password, *args, **kwd):
        return self.session(self, playerName, password, *args, **kwd)
    def resume(self, sessionId):
        return self.session(self, None, None, sessionId = sessionId)

    class session:
        terminateOnGC = False

        def __init__(self, client, playerName, password, sessionId = None, passwordSend = True):
            self.client = client

            if sessionId is None:
                # XXX Todo: session.open method not supported on circle.
                self.sessionId = self.client.server.session.open(playerName)
                self.terminateOnGC = True

                if passwordSend:
                    self.send(password)
            else:
                self.sessionId = sessionId

        def __repr__(self):
            return '<%s %r>' % (self.__class__.__name__, self.sessionId)

        def __del__(self):
            if self.terminateOnGC:
                # XXX Todo: transition to cleanup object
                pass # self.client.server.session.terminate(self.sessionId)

        def send(self, message):
            return self.client.server.session.send(self.sessionId, message)
        def command(self, name, *args):
            # print(f'[phClient.session.command] {name}')
            return self.client.server.session.command(self.sessionId, name, *args)
        def messages(self, n = -1, timeout = None):
            return self.client.server.session.messages(self.sessionId, n, timeout)

        def loginShell(self, password):
            return self.client.server.session.shell.login(self.sessionId, password)

        # Web Adapter Commands
        def interpretAgent(self, script, params = None):
            return self.command('do-interpret-agent', script,
                                *(() if params is None else (params,)))

        def interpretScript(self, script, params = None):
            return self.command('do-interpret-script', script,
                                *(() if params is None else (params,)))

        def interpretAsyncScript(self, script, nr, params = None):
            return self.command('do-interpret-script-async', script, nr,
                                *(() if params is None else (params,)))


@actionable
def stringCall(name, *arguments):
    # Unifies all string states into one string.
    return '%s(%s)' % (name, ', '.join(mapi(str, arguments)))

operator = stringCall('symbolic.operator')
evaluation = stringCall('symbolic')

def symbolic(name):
    return 'symbolic[%r]' % name

class sourceAssemblization:
    'Assemble source nodes.'

    def __call__(self, node):
        (name, children) = node
        # print name

        method = getattr(self, name, self.default)
        return method(name, node, *children)

    def Identifier(self, name, node, identifier):
        # if identifier == 'symbolic':
        #     raise SyntaxError('reserved symbolic keyword')

        return symbolic(identifier)

    def Const(self, name, node, value):
        return value # repr() # Literal.

    def Suite(self, name, node, *children):
        # This may be wrong.
        return str(list(map(self, children)))

    def FunctionCall(self, name, node, method, arguments):
        return evaluation(*list(map(self, [method] + arguments)))

    # def List(self, name, node, identifier):
    #     return

    def FunctionDefinition(self, name, node, functionName, argumentNames, statements):
        pass

    # class Member(BinaryOperation):
    # class Item(BinaryOperation):

    def default(self, name, node, *children):
        return operator(name, *list(map(self, children)))


    # So: these will need to be formed into some kind of
    # assignment virtualization because suites of statements
    # are all formulated into expressions and assignment-
    # based statements shall not be allowed.

    # class Assign(BinaryOperationBase, Statement):
    # class AugmentedAssign(Assign):
    # class PlusAssign(AugmentedAssign):
    # class MinusAssign(AugmentedAssign):
    # class StarAssign(AugmentedAssign):
    # class SlashAssign(AugmentedAssign):
    # class ModAssign(AugmentedAssign):


class phSyntaxObject(dict):
    # The numerical singleton of the lephton language -- or ella-mental interpreter.
    @classmethod
    def FromServer(self, syntax, **kwd):
        return self(syntax, **kwd)

    def __init__(self, syntax, **kwd):
        dict.__init__(self, kwd)
        self.tree = syntax


    @property
    def emulation(self):
        # 'source' is a builtin provided by op.runtime.virtual.objects
        return source.code(self.sourceAssemblization)

    assemblizationClass = sourceAssemblization

    @property
    def sourceAssemblization(self):
        # Return a text format of the syntax tree encoded into an emulation context.
        #
        # The idea is to return a form of the syntax that virtualizes all of the
        # syntax elements that comform to op.compute.syntactic.operational.

        # For now...
        return self.assemblizationClass()(self.tree)


# Install the following object into into operational client.
apiClient.syntaxObject = phSyntaxObject.FromServer


class runtime:
    '''
    # Use as a base class for evaluating compiled emulations.

    client = apiClient(runtime = runtime({'$': library}))
    parallel = action(future.Work, client.run)
    w = parallel(source)

    '''

    def __init__(self, symbols = None, **kwd):
        # Natives.
        if symbols is None:
            symbols = dict()

        symbols.update(kwd)
        self.symbols = symbols


    # Operations.
    def __getitem__(self, name):
        return self.symbols[name]

    def __call__(self, method, *operands, **kwd):
        return method(*operands, **kwd)

    def operator(self, name, *operands, **kwd):
        method = getattr(self, name)
        return method(*operands, **kwd)


    def Addition(self, left, right):
        return left + right


    # Bind to client.
    def run(self, client, source, *args, **kwd):
        code = client.moduleCompile(source).emulation.code
        function = code.function(dict(symbolic = self))

        return function(*args, **kwd)


class nativeSession:
    '''
    itham = 'network/game/session'('runphase.com:2172', 'itham', 'ithams-password')
    task$(_processSessionMessages, session, itham)

    def _processSessionMessages(session, player):
        while true:
            _processMessages(session, player.messages)

    def _processMessages(session, messages):
        for msg in messages:
            act(_processMessage, [session] + msg)

    def _processMessage(session, type):
        msg = arguments().slice(2)

        if type == 'output':
            session.sendln(msg[0])

    _processMessages(session, itham + 'return $.network.stats.users().display()')
    itham('look board')

    '''

    def __init__(self, frame, host, playerName, password, passwordSend = True):
        passwordSend = bool(passwordSend)

        task = frame.task
        task.checkAccessSystem(['system:network', 'https', host, 'rpc'], 'write')
        client = apiClient(None, None, host = host) # , protocol = 'http')

        task -= frame

        # print(f'[phClient.session] {client} {task} {self._pool}')
        # debugOn()


        @self._pool.task(task)
        def connect(_):
            from xmlrpc.client import Fault

            print(f'[phClient.session.connect] START: {client} {playerName}')

            try: self._session = client.login(playerName, password, passwordSend = passwordSend)
            except Fault as e:
                # XXX Todo: hide system traceback. :skip:
                raise e

            print(f'[phClient.session.connect] DONE: {self}')
            return self

    def _action(self, function, *args, **kwd):
        from stuphos.kernel import vmCurrentTask
        task = vmCurrentTask()
        task -= task.frame

        @self._pool.task(task)
        def action(_):
            # from xmlrpc.client import Fault

            try: return function(*args, **kwd)
            # except Fault as e:
            except Exception as e:
                # XXX Todo: hide system traceback. :skip:
                raise e

            # try: result = function(*args, **kwd)
            # except Exception as e:
            #     debugOn()
            #     raise e

            # debugOn()
            # return result

    def send(self, message):
        return self._action(self._session.send, message)
    def command(self, name, *args):
        return self._action(self._session.command, name, *args)

    __call__ = send

    # @property
    def messages(self):
        return self._action(self._session.messages, timeout = 30)

    def loginShell(self, password):
        return self._action(self._session.loginShell, password)

    def interpretAgent(self, script, params = None):
        return self._action(self._session.interpretAgent, script, params = params)
    def interpretScript(self, script, params = None):
        return self._action(self._session.interpretScript, script, params = params)

    def interpretAsyncScript(self, script, nr, params = None):
        return self._action(self._session.interpretAsyncScript, script, nr, params = params)


    def __add__(self, script):
        return self.interpretScript(script)
    __lshift__ = __add__


    @property
    def _lookbackSession(self):
        # XXX port
        if self._session.client.host in ['localhost', '127.0.0.1']:
            mgr = runtime[runtime.Web.Adapter.SessionManager]
            return mgr.getSessionByIdEx(None, self._session.sessionId)



'Implements the events invoked from Crash_ in objsave.'

def deleteRentFile(name, info, filename):
    world.corelog('(DeletingRent:%s:%s)' % (name, filename))
    # showRentInfo(info)

from new import instancemethod as bind

class RentManagement:
    msgfmt = '(PC) %(Name)s %(Code)s%(Action)s from %(Filename)r at [%(Timestamp)s]'

    def loadNotifiers(self, a):
        'Load the rent handlers.'
        def logAction(action, player, info, filename):
            info = self.unpackRentInfo(info)
            info = {'Name':player.name, 'Code':info[0], 'Action':action,
                    'Timestamp':info[1], 'Filename':filename}

            world.corelog(self.msgfmt % info)

        a.rentStart       = lambda *args:logAction('SaveStart',    *args)
        a.rentComplete    = lambda *args:logAction('SaveComplete', *args)
        a.unrentStart     = lambda *args:logAction('LoadStart',    *args)
        a.unrentComplete  = lambda *args:logAction('LoadComplete', *args)

    __init__ = loadNotifiers

    from struct import unpack
    unpack = staticmethod(unpack)

    # time, code, perdiem, gold, acct, nitems, <8 spares>
    structfmt_rent_info = '14i'
    rent_info_member_names = ('time', 'code', 'perdiem', 'gold', 'acct', 'nitems')

    rent_codes = { 1: 'Crash', 2: 'Rented', 4: 'Forced', 5: 'Timed-Out' }

    from time import ctime
    ctime = staticmethod(ctime)

    def unpackRentInfo(self, info):
        'Unpack the rent_info buffer and parse rent-code, time.'
        data = self.unpack(self.structfmt_rent_info, str(info))

        # 2-tuple
        return (self.rent_codes.get(data[1]) or ('#%d ' % data[1]), self.ctime(data[0]))

        # [for (x, n) in zip(data[:6], self.rent_info_member_names)]




# Note: this is not an an interface to the operating system, this is a system in itself
# that represents a hosting compartment for the application.  (It is used by framework).

# An example a pure python command-line bootstrap:
# --
# def runCore(argv = None):
#     from stuphos.system.core import Core
#     from stuphos.system.cli import getCmdln

#     # These need to be provided...
#     # from stuphos import stuphlib
#     # from game import world

#     # But only for world-booting purposes which don't exist without a directory:
#     # def bootWorld(self, options, stuphlib, worldModule):
#     #     if options.world_dir:

#     stuphlib = world = None

#     # Run application -- might want to provide at least the headless option.
#     Core(getCmdln(argv),
#          stuphlib, world,
#          globalize = True)


# Aspects -- this is game-support, not the actual mud package heartbeat.
__all__ = ['Heartbeat', 'Game', 'mudlog', 'core']

class Heartbeat:
    class Task:
        def perform(self, engine):
            pass

        # Task-Specific.
        def activate(self, engine):
            pass
        def deactivate(self, engine):
            pass

        # Engine-Specific.
        def engine_started(self, engine):
            pass
        def engine_stopped(self, engine):
            pass

    def __init__(self):
        self.tasks = []

    def run(self, optimisticTermination = True):
        # Game-loop.
        try:
            self.start()
            self.perform_start()
            while self.is_running():
                self.perform_cycle()

                # debugOn()
                if not optimisticTermination and not self.alive():
                    # performance decision -- END
                    self.stop()

        finally:
            self.perform_stop()

    def perform_cycle(self):
        self.dispatch_operation('perform')

    # Task Interface
    def dispatch_operation(self, name):
        for task in self.tasks:
            op = getattr(task, name, None)
            if callable(op):
                # print "Dispatching '%s' on %r" % (name, task)
                op(self)

    def is_running(self):
        return getattr(self, '_running_state', False)
    def set_running(self, state = True):
        previous_state = self.is_running()
        state = bool(state)
        setattr(self, '_running_state', state)

    def perform_start(self):
        self.dispatch_operation('engine_started')
    def perform_stop(self):
        self.dispatch_operation('engine_stopped')

    def start(self):
        self.set_running(True)
    def stop(self):
        self.set_running(False)

    def alive(self):
        if bool(self.tasks):
            from world import heartbeat as vm
            for x in vm.tasks:
                return True

        return False

    def activate(self, task):
        if task not in self.tasks:
            self.tasks.append(task)
            task.activate(self)
        return self
    def deactivate(self, task):
        if task in self.tasks:
            self.tasks.remove(task)
            task.deactivate(self)
        return self

    __iadd__ = activate
    __isub__ = deactivate

class Game(Heartbeat.Task):
    # Internal.
    import queue
    import traceback

    # System Task -- Event Queue
    def __init__(self, timeout):
        self.event_q = self.queue.Queue()
        self.timeout = timeout

    def activate(self, engine):
        engine.event = self

    def call(self, event, *params, **kwd):
        self.event_q.put((event, params, kwd))

    def __iadd__(self, event):
        self.call(event)
        return self

    def perform(self, engine):
        timeout = self.timeout

        while True:
            if timeout is not False:
                (event, params, kwd) = self.event_q.get(timeout)
            else:
                try: (event, params, kwd) = self.event_q.get_nowait()
                except self.queue.Empty:
                    timeout = self.getDefaultTimeout()
                    if not timeout:
                        break

                    continue

            try: event(*params, **kwd)
            except SystemExit:
                break
            except:
                # Soft Fail.
                self.traceback.print_exc()

    def getDefaultTimeout(self):
        return False # 0.1

class Interpreter(Heartbeat.Task):
    # Command Dispatch.
    pass

class ManualDelay(Heartbeat.Task):
	def __init__(self, core):
		self.ev = Event()
		core += self
		core.switch = self

	def perform(self, engine):
		print('Waiting on', self.ev, '...')
		self.ev.wait()
		print('Done waiting on ', self.ev, '.')

	def signal(self):
		print('Sending signal to', self.en, '...')
		self.ev.set()

	__call__ = signal

class EventBridge:
    # A similar, unused symbol is defined in runtime/events.
    class Calling:
        def __init__(self, name):
            self.name = name
        def __call__(self, *args, **kwd):
            try: from stuphos.system.api import game
            except ImportError:
                from stuphos.system import game

            bridgeModule = game.bridgeModule()
            if bridgeModule:
                function = getattr(bridgeModule, self.name, None)
                if callable(function):
                    return function(*args, **kwd)

        def __repr__(self):
            return 'Calling: %r' % self.name
        __str__ = __repr__

    def __setattr__(self, name, value):
        self.__dict__[name] = self.Calling(value)

# The default bridge.  Application may override this.
import builtins as builtin
if not hasattr(builtin, 'StuphMUD'):
    builtin.StuphMUD = EventBridge()

# Todo: Export this constant.
LVL_IMMORT = 106

def mudlog(message, level = LVL_IMMORT, typeName = "Normal"):
    from world.player import players

    message = str(message)
    level = int(level)
    typeName = typeName and str(typeName) or "Normal"

    # Todo: Colorization.
    msg = '[ %s ]' % message
    for peer in players():
        # Todo: If not writing/mailing/creating and is playing.
        # Todo: Type-level check.
        if peer.avatar is not None and level <= peer.avatar.level:
            print(msg, file=peer)

def boot_time():
    from builtins import core
    return core.bootStartTime



#!/usr/local/bin/python -i
"""CircleMUD object database file interpreter.

Version: %(__version__)f

Class: ObjectReader
	constructor - first argument is either an open file handle,
		a string path name for a file to open, or a zone number
		for loading from a builtin library path

	parseFile() - initiate process to read entire file.  Calls
		startOfRecords(), newRecord(), recordField(), recordDescription(),
		recordAffection() and endOfRecords().
		Override these methods on subclass to implement your own
		records processing.

	result field - dictionary holding entire parsed-file database

"""
# Fraun Dec 24th 2005 - Converted to use RecordReader.

__version__ = 0.3
# __all__ = ['ObjectReader', 'ExtraDescription']

def usage():return __doc__%{'__version__':__version__}

from .dblib import *

class ExtraDescription:
	def __init__(self, keyword, text):
		self.keyword=keyword
		self.text=text

	#def __repr__(self):pass

class Affection:
	def __init__(self, location, modifier):
		self.location=location
		self.modifier=modifier

	#def __repr__(self):pass

class ObjectReader(RecordReader):
	def parseRecord(self):
		"""Load record area into logical Python object

		file argument is an open file handle

		Format of textually-defined object prototype:

		<Name String (keywords)>~<NL>
		<Short Description>~<NL>
		<Long Room Description (multiline)><NL>
		~<NL>
		<Action Description (multiline)><NL>
		~<NL>
		<Type as Integer Code> <ASCII-flag for Extra Bits> <ASCII-flag for Wear Bits>
			<ASCII-flag for Anti Bits><NL>
		<Integer Object Value 1> <Value 2> <Value 3> <Value 4><NL>
		<Integer Weight> <Cost> <Rent><NL>

			Extra Description/Affection/Special declarations Follow:
		E
		<Extra Description keywords>~<NL>
		<Extra description text (multiline)><NL>
		~<NL>
		A
		<Integer Affect Location Code> <Modifier Code><NL>
		S

			Singular Special (espec) Fields Follow:
		Timer: -1 to 100
		Trap: <Integer code>

		"""

		# Pass each field read to handler function
		self.recordField('name', self.tildeString('name'))
		self.recordField('shortdescr', self.tildeString('shortdescr'))
		self.recordField('longdescr', self.tildeString('longdescr'))
		self.recordField('actiondescr', self.tildeString('actiondescr'))

		type, extra, wear, anti = self.readLine().split()
		self.recordField('type', int(type))
		self.recordField('extraflags', ASCIIFlagDecode(extra))
		self.recordField('wearflags', ASCIIFlagDecode(wear))
		self.recordField('antiflags', ASCIIFlagDecode(anti))

		self.recordField('vals', [int(v) for v in self.readLine().split()])

		weight, cost, rent = self.readLine().split()
		self.recordField('weight', int(weight))
		self.recordField('cost', int(cost))
		self.recordField('minlevel', int(rent)) # Rent is minlevel..

##		timer=trap=0

##		extraDescr=[]
##		affections=[]

		line=self.readLine()

		while line!='':
			c=line[0]

			if c=='$': # detected end of records
				#print '} End Last Object'
				return

			if c=='#': # start of next record
				#print '} End Object, Next'
				try: return int(line[1:])
				except ValueError:
					raise FormatError('Expected #<vnum> of next record; got %r' % line[1:])

			if c=='E':
				self.objectDescription(
					self.tildeString('exdesc keywords'),
					self.tildeString('exdesc descr'))

			else:
				# This is done here because these blocks all uses additional lines
				line = self.readLine()

				assert line != '', FormatError("EOF expecting data for '%c'" % c)

				if c == 'A':
					location, modifier = line.split()

					self.objectAffection(location, modifier)

				elif c == 'S':
					while True:
						# Fraun Mar 1st 2006 - Note the necessary heading ' ' involved
						if line[:1] == '$':break # End of 'S' section.
						elif line[:1] == '#':
							# This isn't part of of parse_espec, but I wonder about old formats...
							# Todo: Warn.
							try: return int(line[1:])
							except ValueError:
								raise FormatError('Expected #<vnum> of next record; got %r' % line[1:])

						# Eww, but at least it's lowercase.
						elif line[:7] == 'Timer: ':self.recordField('timer', int(line[7:]))
						elif line[:6] == 'Trap: ':self.recordField('trap', int(line[6:]))
						elif line[:10] == 'SpecProc: ':self.recordField('specproc', line[10:])
						else:raise FormatError('Unknown S-type line "%r" (object #%d; line %d)' % (line, self.getVirtualNumber(), self.lineno))

						line = self.readLine()
						if line == '':
							return self.EOFNoDollar()

				else:
					raise FormatError("Unknown E/A/S-type '%r' (object #%d; line %d)" % (makePrintable(c), self.getVirtualNumber(), self.lineno))

			line=self.readLine()

		# No $ at terminating records at end of file
		#print '} End Object -- EOFNoDollar'
		return self.EOFNoDollar()

	# This is called for parsed EXTRA descriptions.  Other string descriptions
	# are parsed as recordFields
	def objectDescription(self, keyword, text):
		r = self.getCurrentRecord()

		if 'extradescr' in r:
			# ExtraDescription(keyword, text)
			r['extradescr'][keyword] = text
		else:
			# ExtraDescription(keyword, text)
			r['extradescr'] = {keyword:text}

	# Object affection data
	def objectAffection(self, location, modifier):
		aff = Affection(location, modifier)
		r = self.getCurrentRecord()

		if 'affections' in r:
			r['affections'].append(aff)
		else:
			r['affections'] = [aff]

## SQLObject representation.
DEFAULT_ITEM_DBFL = 'lib/etc/items.db'

def openItemTables(dbpath):
	from .sql import openDatabaseTables
	return openDatabaseTables \
		(dbpath, 'StuphItemRecord', schema_package = 'sql.item')

def convert_item_row(vnum, item, columns):
	'Convert Object Record into a mapping suitable for SQLObject row representation.'
	from .sql import getStuphZoneRecordIDByZoneVnum

	row = {'vnum' : vnum}
	for n in columns:
		if n in item:
			row[n] = item[n]

		# (S)pecial .obj file record section
		elif n in ('timer', 'trap'):
			row[n] = 0
		elif n in ('specproc',):
			row[n] = ''

	# Join with separate tables for extra-descriptions.  values? affections??
	row['vals']       = buffer(str(row.get('vals',  [0,0,0,0])))
	row['extradesc']  = buffer(str(row.get('extradescr',   [])))
	row['affections'] = buffer(str(row.get('affections',   [])))

	# Resolve this.
	row['stuphZoneRecordID'] = getStuphZoneRecordIDByZoneVnum(vnum / 100)

	return row

def loadStuphItemRecords(dbpath = DEFAULT_ITEM_DBFL, populate_with = None):
	StuphItemRecord = openItemTables(dbpath)[0]

	if populate_with:
		def iterate_population(p):
			if type(p) is list:
				for x in p:
					for y in iterate_population(x):
						yield y
			elif type(p) is str:
				for x in ObjectReader(p).parseFile().consumeRecordSet().items():
					yield x

		columns = list(StuphItemRecord.sqlmeta.columns.keys())

		for (vnum, item) in iterate_population(populate_with):
			StuphItemRecord(**convert_item_row(vnum, item, columns))

	return StuphItemRecord

def test():
	reader=ObjectReader('lib/0.obj')
	reader.parseFile()

	return reader.consumeRecordSet()

def main(opts = []):
	if '-test' in opts:
		print(test())
	else:
		n = len (opts)
		i = 0

		populate_with = []
		dbfile = DEFAULT_ITEM_DBFL

		while i < n:
			a = opts[i]

			if a in ('-new', '-new-db', '-newdb'):
				dbfile = new_database_file()
			elif a in ('-db', '-dbfile', '-dbpath'):
				i += 1
				dbfile = opts[i]
			else:
				populate_with.append(a)

			i += 1

		return loadStuphItemRecords(dbfile, populate_with = populate_with)

if __name__ == '__main__':
	from sys import argv

	if '-debug' in argv:
		from pdb import runcall
		main_app = main

		def main(*args):
			return runcall(main_app, *args)

		argv.remove('-debug')

	I = main(argv[1:])



"""
Implement an interpreter-based OLC, which loads a syslog (by virtual FS) as an input source,
subjecting each line to a match against a pattern database, which matches and processes messages
based on their format.  The menu-based editor is used to process lines against existing templates,
and to allow provision of new templates for processed lines.

Process lines come from streamed sources: they can be from a file, possibly a re-scan of the entire
syslog, or, as a branched processing intercept from the mudlog programmable bridge event.  The editor
can then be used to edit a certain new batch of incoming lines that have not been matched yet.

The matched lines database associates with handlers that are called in the final target processing
of ongoing mudlog events.  It is the responsibility of the scheduler executing these handlers towards
unifying them within the mud whole.

---
Implementation:
        
"""

from stuphos.etc.tools.strings import compactify, safeAttr, isYesValue
from stuphos.runtime.architecture.api import NoAccessException

import re, errno, pickle, traceback,pprint
from os import popen

try: import readline as rl
except ImportError: rl = None

# Generic serializable pattern construct.
class pattern:
    def __init__(self, pattern):
        self.pattern = pattern

    def __getstate__(self):
        return self.pattern

    def __setstate__(self, pattern):
        self.pattern = pattern

    def __str__(self):
        return str(self.pattern)

    __reduce__ = __str__

class compile_pattern(pattern):
    'Implements regular-expression match.'
    TYPE = 'compiled'

    def __init__(self, pattern):
        self.pattern = pattern
        # self._compiled = re.compile(pattern)
        self.match = re.compile(pattern).match

    __setstate__ = __init__

class static_pattern(pattern):
    'Implements a simple equality match.'
    TYPE = 'static'

    def match(self, line):
        return line == self.pattern


logfile_timefmt_header = 'mon day 24h:m:s year     :: '

class PatternMatch:
    'Serializable database pattern-matching algorithm.'

    patternCollection = list # set

    def __init__(self, patterns = None):
        self.patterns = self.patternCollection(patterns) \
                            if patterns else self.patternCollection()
        self.lineno = 0
        self.processMatch = None

    def __getstate__(self):
        return (self.patterns, self.lineno)

    def __setstate__(self, state):
        (patterns, lineno) = state

        self.patterns = self.patternCollection(patterns)
        self.lineno = lineno

    # The algorithm, with callbacks:
    def matchlines(self, source):
        'Scan iteratable line source and generate full-stage pattern recognition processing.'

        for line in self.iterateLines(source):
            self.startMatch(line)

            result = False
            for pat in self.patterns:
                match = pat.match(line)
                if match:
                    result = True
                    yield (match, pat, line, self.lineno)

            if not result:
                yield (None, None, line, self.lineno)

            self.endMatch(line)

    def iterateLines(self, source):
        for line in source:
            self.lineno += 1

            # yield line
            # self.timestamp = strptime(logfile_timefmt_header, line[:28])
            self.timestamp = line[:28]
            yield line[28:]

    def startMatch(self, source):
        pass

    def endMatch(self, source):
        pass

    def addPattern(self, pattern):
        s = self.patterns
        if type(s) is list and pattern not in s:
            s.append(pattern)
        elif type(s) is set:
            s.add(pattern)

    def getPattern(self, index):
        return self.patterns[index] if type(self.patterns) is list else None

    def setPattern(self, index, pattern):
        if type(self.patterns) is list:
            self.patterns[index] = pattern

    __getitem__ = getPattern
    __setitem__ = setPattern

class InteractiveMatch(PatternMatch):
    'Interactive pattern match shell used for editing the syslog templates.'
    # High-level interface for pattern-matching database.

    def noMatch(self, line, lineno):
        'Interactive prompt when no match is found for a line.'

        while 1:
            resp = self.doPrompt(line).strip()
            if resp == '':
                break

            lresp = resp.lower()

            if resp == '=':
                # Add a static pattern for this line.
                self.addPattern(static_pattern(line))
                print('Static Pattern added for:\n  %r' % line)
                break

            elif lresp in ('-h', '-he', '-hel', '-help', '--h', '--he', '--hel', '--help'):
                self.doHelp(lresp)

            elif lresp in ('-s', '-sh', '-sho', '-show', '--s', '--sh', '--sho', '--show'):
                self.doShow(lresp)

            elif lresp in ('-sa', '-sav', '-save', '--sa', '--sav', '--save'):
                self.save()
                n = len(self.patterns)
                print('%d pattern%s saved.' % (n, 's' if n != 1 else ''))

            elif '-populate'.startswith(lresp):
                self.populatePatternHistory()

            elif '-replace'.startswith(lresp):
                pass
            elif '-up'.startswith(lresp):
                pass
            elif '-down'.startswith(lresp):
                pass
            elif '-back'.startswith(lresp):
                pass
            elif '-forward'.startswith(lresp):
                pass
            elif '-select'.startswith(lresp):
                pass

            else:
                try:
                    pat = compile_pattern(resp)
                except:
                    traceback.print_exc()
                    continue

                if not pat.match(line):
                    print('line does not match:', resp)
                    print('  ', line)
                else:
                    self.addPattern(pat)
                    break

    def doPrompt(self, line):
        return input('Provide a template match for:\n   %r\n? ' % line)

    def doHelp(self, command):
        self.page(re.__doc__)

    def doShow(self, command):
        s = self.patterns

        if type(s) is list:
            from io import StringIO
            b = StringIO()

            for x in range(len(s)):
                print('#%3d :' % x, str(s[x]).rstrip(), file=b)

            self.page(b.getvalue())
        else:
            self.page('\r\n'.join(str(pat).rstrip() for pat in s))

    def page(self, msg, columnize = False):
        popen('less' if not columnize else 'column|less', 'w').write(msg)

    def populatePatternHistory(self):
        s = self.patterns
        if type(s) is list:
            x = len(s)
            def get(i):
                return str(s[i]).rstrip()

            n = rl.get_current_history_length()
            if n > x:
                n = x

            for i in range(n):
                rl.replace_history_item(i, get(i))

            if x > n:
                for i in range(n, x):
                    rl.add_history(get(i))

    def interactive(self, source = None, notifyMatch = None):
        if source is None:
            source = getattr(self, 'source', None)

        assert source
        self.source = source
        if not notifyMatch:
            notifyMatch = getattr(self, 'notifyMatch', None)

        try:
            for (match, pat, line, lineno) in self.matchlines(source):
                if match:
                    notifyMatch(match, pat, line, lineno)
                else:
                    self.noMatch(line, lineno)

        except EOFError:
            pass

    resume = interactive

    def notifyMatch(self, match, pattern, line, lineno):
        if hasattr(pattern, 'handleMatch'):
            return pattern.handleMatch(match, line, lineno)

        print('Found (line #%d): %r' % (lineno, line))
        if type(match) is not type(True):
            print('  ', str(pattern).rstrip())
            print('  ', ', '.join([_f for _f in match.groups() if _f]))

            gd = match.groupdict()
            if gd:
                pprint.pprint(gd)

    def save(self, file = None):
        pickle.dump(self, open(str(file or self.templates_file), 'w'))

    def export(self, filename):
        out = open(filename, 'wt')
        type = None
        nr = 0

        for pattern in self.patterns:
            if type != pattern.TYPE:
                type = pattern.TYPE
                if type is not None:
                    print(file=out)

                print('[%s]' % type.capitalize(), file=out)

            # XXX :skip: careful:
            pattern = str(pattern).rstrip()

            print('%d: %s'  % (nr, pattern), file=out)
            nr += 1

        out.flush()
        out.close()

def loadPatternMatcher(templates_file):
    patternMatcher = InteractiveMatch
    try:
        if templates_file.endswith('.pkl'):
            # Deserialize.
            patmatch = pickle.load(open(templates_file))
        else:
            # Load templates from regular file -- constructing regexprs.
            patmatch = patternMatcher(compile_pattern(pattern) for pattern in open(templates_file))

        patmatch.templates_file = templates_file

    # Else default.
    except:
        from sys import exc_info
        (etype, value, tb) = exc_info()
        print('[syslog pattern loader] %s: %s' % (etype.__name__, value))

        patmatch = patternMatcher()
        patmatch.templates_file = templates_file

    return patmatch

def loadProcessor(name):
    if name:
        name = name.split('.')
        if len(name) == 1:
            return globals().get(name[0])

        try: module = __import__('.'.join(name[:-1]), globals(), locals(), [''])
        except ImportError: pass
        else: return getattr(module, name[-1], None)

def ignoreMatch(*args):
    pass

## Scripted Main.
DEFAULT_PATTERNS = 'stuph-log-patterns.pkl'

def main(argv = None):
    # Parse command line.
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('--database', '--db')
    parser.add_option('-g', '--debug', action = 'store_true')
    parser.add_option('-i', '--input-file', '--source')
    parser.add_option('-s', '--show-patterns', action = 'store_true')
    parser.add_option('-p', '--processor')
    parser.add_option('-x', '--export')
    parser.add_option('--import')
    (options, args) = parser.parse_args()

    if options.input_file:
        source = open(options.input_file)
    elif args:
        assert len(args) == 1
        source = open(args[0])

    templates_file = options.database or DEFAULT_PATTERNS

    if options.debug:
        from pdb import set_trace
        set_trace()

    # Load the pattern matcher.
    patmatch = loadPatternMatcher(templates_file)
    if options.export:
        patmatch.export(options.export)
    else:
        processMatch = loadProcessor(options.processor)

        # Interact.
        if options.show_patterns:
            for pattern in patmatch.patterns:
                print(pattern.pattern)
        else:
            patmatch.interactive(source, processMatch)

if __name__ == '__main__':
    main()

USAGE = \
'''
This subcommand will compare a collection of syslogs against a category
of patterns (use --show=all or --show=static|compiled to get a list of
categories).

Use --show-paths to see a list of paths that the scanner searches, and
--category=<category seen with --show> to do the actual search.
'''

from optparse import OptionParser
syslogCommandParser = OptionParser(USAGE)
syslogCommandParser.add_option('-c', '--category')
syslogCommandParser.add_option('-s', '--show-categories', '--show')
syslogCommandParser.add_option('--show-paths', action = 'store_true')

def parseScanSyslogCmdln(args):
    from ..etc import parseOptionsOverSystem
    return parseOptionsOverSystem(syslogCommandParser, args)

def getScanSyslogPaths(section):
    from glob import glob

    import re
    pathpat = re.compile('path(?:\.(\d+))?').match

    for opt in section.options():
        if pathpat(opt):
            for p in glob(section.get(opt)):
                yield p

class Instruction:
    def __init__(self, command):
        self.command = command

    context = []

    class Context(list):
        def __init__(self, name):
            self.name = name

    @classmethod
    def isState(self, line):
        if line:
            if line[0] == '[' and line[-1] == ']':
                line = line[1:-1].strip()
                i = line.find(' ')
                if i >= 0:
                    state = line[i:]

                state = line.lstrip()

                if state.lower() == 'state:':
                    line = line[len(state):].lstrip()
                    if line[0] == '@':
                        def i():
                            yield self(line[1:])

                        # Return a non-None generator on state match.
                        return i()

    def processState(self, buffer):
        args = self.command.split()
        cmd = args[0]

        if cmd == 'open':
            self.context.append(self.Context(' '.join(args[1:])))

        elif cmd == 'close':
            try: 
                for match in buffer:
                    yield '  ' * len(self.context) + match

            finally:
                del buffer[:]
                self.context.pop()

    def endState(self):
        # Stream the buffer.
        for match in self:
            yield match


    @classmethod
    def build(self, scan):
        # Process state indentation.
        self.context.append(self.Context(''))

        for match in scan(self.isState):
            if isinstance(match, self):
                for match in match.processState(self.context[-1]):
                    yield match
            else:
                self.context[-1].append(match)

        # Unwind any unclosed states.
        for ctx in range(len(self.context)):
            for match in self.context[ctx].endState():
                yield match


# These should be: complex and basic
KNOWN_PATTERN_TYPES = ['Compiled', 'Static']
MINIMUM_LEVEL = 115
def doScanSyslog(peer, cmd, argstr):
    # todo: caching of patterns object, syslogs (pre-match index)
    # to disk.. reloaded via md5 checksum change detection.
    args = argstr.split() if argstr else []
    if not args or args[0].lower() != 'scan':
        return False

    del args[0]
    if peer.avatar and peer.avatar.level >= MINIMUM_LEVEL:
        try:
            (options, args) = parseScanSyslogCmdln(args)

            from stuphmud.server import getSection
            section = getSection('Syslog')

            syslog_paths = getScanSyslogPaths(section)
            patterns = section.get('patterns')

            if options.show_paths:
                def showPaths():
                    yield '&ySyslog Paths:&N'
                    yield '============='
                    yield ''

                    for path in syslog_paths:
                        yield path

                peer.page_string('\n'.join(showPaths()) + '\n')
                raise SystemExit

            if not patterns:
                print('Not syslog patterns defined!', file=peer)
                raise SystemExit

            # Load appropriate pattern category:
            # todo: actually provide categorical combinations.
            from configparser import ConfigParser, NoOptionError, NoSectionError
            cfg = ConfigParser()
            cfg.readfp(open(patterns), patterns)

            show = options.show_categories
            if show:
                # todo: wildcard--search patterns themselves!
                show = show.lower()

                def showSections():
                    yield '&ySyslog Categories:&N'
                    yield '=================='
                    yield ''

                    for section in cfg.sections():
                        if section not in KNOWN_PATTERN_TYPES:
                            continue

                        lsection = section.lower()
                        if show in ['any', 'all'] or show == lsection:
                            yield str(section)
                            for opt in cfg.options(section):
                                yield '  [%-5s] %s' % (opt, cfg.get(section, opt))

                            yield ''

                peer.page_string('\n'.join(showSections()) + '\n')
                raise SystemExit

            if not options.category:
                print('Category is required.', file=peer)
                print(file=peer)

                parseScanSyslogCmdln(['-h'])
                raise SystemExit

            # Check aliases, first:
            category_id = options.category
            try: category_id = cfg.get('Aliases', category_id)
            except (NoOptionError, NoSectionError): pass

            for type in KNOWN_PATTERN_TYPES:
                try: pattern = cfg.get(type, category_id)
                except (NoOptionError, NoSectionError): pass
                else:
                    if type == 'Compiled':
                        category = compile_pattern(pattern)
                    elif type == 'Static':
                        category = static_pattern(pattern)

                    # Get presentation format.
                    try: category_header = cfg.get('Presentation', '%s.header' % category_id, raw = True)
                    except (NoOptionError, NoSectionError): category_header = None

                    try: category_format = cfg.get('Presentation', '%s.format' % category_id, raw = True)
                    except (NoOptionError, NoSectionError): category_format = None

                    break
            else:
                print('Unknown category: %r' % category_id, file=peer)
                raise SystemExit

            # Do processing.
            def scanSyslog(isState):
                yield '&ySyslog Contents:&N'
                yield '================'
                yield str(category)
                yield ''

                for path in syslog_paths:
                    try: stream = open(path)
                    except IOError:
                        # todo: test for non (is-directory or enoent)
                        continue

                    first_in_file = True
                    linenr = 0
                    for line in stream:
                        linenr += 1

                        # XXX :skip: Careful:
                        line = line.rstrip()
                        line = line[28:]

                        match = isState(line)
                        if match is not None:
                            # Process state.
                            for match in match:
                                yield match

                            # Move to next incoming stream line.
                            continue

                        match = category.match(line)

                        if match is True:
                            if first_in_file:
                                yield '%s:' % path
                                first_in_file = False

                            yield '  %-5d: %s' % (linenr, line)

                        elif match not in (False, None):
                            if first_in_file:
                                yield '%s:' % path
                                if category_header:
                                    yield '         %s' % category_header
                                    yield '         %s' % ('-' * len(category_header))

                                first_in_file = False

                            groups = match.groupdict()
                            if not groups:
                                groups = match.groups()

                            if category_format:
                                yield '  %-5d: %s' % (linenr, category_format % groups)
                            else:
                                yield '  %-5d: %s' % (linenr, groups)

                    if not first_in_file:
                        yield ''

            # In another thread (how might this affection the active syslog??)
            def process(page_string):
                from stuphos import enqueueHeartbeatTask
                enqueueHeartbeatTask(page_string, '\r\n'.join(Instruction.build(scanSyslog)))

            from _thread import start_new_thread as nth
            nth(process, (peer.page_string,))

        except SystemExit:
            pass

        return True

try: from stuphmud.server.player import ACMD
except ImportError: pass
else: ACMD('syslog*')(doScanSyslog)


from stuphos.runtime.facilities import Facility
from stuphos.db import dbCore
from stuphos.kernel import Native
from stuphos.etc.tools.timing import date as now
from stuphos.etc.tools.logs import tracebackString, exceptionHeader
from stuphos import getConfig

from queue import Queue
from datetime import datetime
from time import time as now

class Journal(Facility):
    NAME = 'System::Journal'

    def __init__(self):
         # [mud/runtime/core.py:225] installServices
         #   Logging.get(create = True)

        # self.taskWaiters = dict()
        self.dbNamespace = getConfig('database', 'Logging')

        self.queue = Queue()

        if self.init() is not False:
            self.startStream()

    def init(self):
        from stuphos.db.orm import Log, createSQLObjectTable
        from stuphos.db import dbCore

        with dbCore.hubThread(self.dbNamespace):
            createSQLObjectTable(Log)


    def processStream(self, write, get):
        # print(f'[vsz] starting stream: {psOpGameVsz()}')
        self.running = True
        try:
            while True:
                job = None
                try:
                    # print(f'[vsz] getting journal job: {psOpGameVsz()}')
                    job = get()
                    write(job)

                except StopIteration:
                    break

                except Exception as e:
                    if job is not None:
                        import json

                        from sys import exc_info
                        (etype, value, tb) = exc_info()

                        try: print(json.dumps(job, indent = 1))
                        except Exception as r:
                            print('reporting job data [%s]: %s' % (r.__class__.__name__, r))
                            print('...%s.%s: %s' % (e.__class__.__module__, e.__class__.__name__, e))

                            from traceback import print_exception
                            print(job)
                            print_exception(etype, value, tb)

                            continue

                    from stuphos import logException
                    logException(traceback = True)
        finally:
            self.running = False

    def startStream(self):
        from stuphos.etc import nth
        nth(self.processStream, self.writeLogTraceback,
            self.queue.get)

    def stopStream(self):
        self.queue.put(StopIteration)

    def __registry_delete__(self):
        self.stopStream()

    def logTraceback(self, task, traceback, exception = None):
        # try: waiter = self.taskWaiters[task.taskName]
        # except KeyError:
        #     pass
        # else:
        #     # Just dispatch the log (don't store)
        #     return waiter(self, task, traceback)

        self.logTracebackDB(task, traceback, exception = exception)

    def writeLogTraceback(self, msg):
        if isinstance(msg, StopIteration) or msg is StopIteration:
            raise msg

        # return # XXX :skip: absorb all until we can migrate Log

        if isinstance(msg, dict):
            source = msg['source']
            type = msg['type']
            content = msg['content']
            timestamp = msg['timestamp']
        else:
            (source, type, content, timestamp) = msg

        from stuphos.db.orm import Log
        from stuphos.db import dbCore

        # print(f'[vsz] writing log: {psOpGameVsz()}')

        with dbCore.hubThread(self.dbNamespace):
            try: Log(source = source, type = type,
                     content = content,
                     timestamp = timestamp).sync()

            except sqlite.module.dberrors.OperationalError as e:
                # Contention?
                logOperation(f'Log ERROR [{self.dbNamespace}]: {e.__class__.__name__}: {e}')

            except ImportError as e:
                # Yep.
                import re
                pattern = re.compile(r'^Failed to import ([_a-zA-Z0-9]+) because the import lock is held by another thread.$').match
                if pattern(str(e)) is not None:
                    # Because this is intermittant, we debug it...
                    print(e)

                    if isYesValue(configuration.Logging.debug_import_lock_error):
                        debugOn()
                else:
                    raise e

        # print(f'[vsz] wrote log: {psOpGameVsz()}')

    def copyToDB(self, fromNS, toNS):
        from stuphos.db.orm import Log
        from stuphos.db import dbCore

        with dbCore.hubThread(fromNS):
            # Copy to memory.
            data = [(e.source, e.type, e.content, e.timestamp)
                    for e in Log.select()]

        with dbCore.hubThread(toNS):
            try: Log.createTable()
            except Exception as e:
                print(f'{toNS}: {e}')

            for (source, type, content, timestamp) in data:
                # Write to database.
                Log(source = source, type = type,
                    content = content,
                    timestamp = timestamp).sync()

        return len(data)


    def __add__(self, msg):
        self.queue.put(msg)
        return self
    __iadd__ = __add__

    def logTracebackDB(self, task, traceback, exception = None):
        traceback = '\n'.join(task.formatTraceback(traceback))
        if exception is not None:
            (etype, value, tb) = exception
            try: header = exceptionHeader(etype, value)
            except NoAccessException:
                try: header = f'{etype.__name__}: {repr(value)}'
                except NoAccessException:
                    header = f'{etype.__name__}: id #{id(value)}'

            traceback += '\n%s:\n%s\n' % (header, tracebackString(tb))

        self += (task.name, 'traceback', traceback,
                 datetime.fromtimestamp(now()))

    # todo: differentiate between logs: traceback, other?
    def getLogs(self, task = None, n = -1, taskName = None, preserve = False, type = 'traceback'):
        from stuphos.db.orm import Log
        from stuphos.db import dbCore

        if task is None:
            assert taskName
        else:
            taskName = task.taskName

        assert n
        with dbCore.hubThread(self.dbNamespace):
            for log in Log.selectBy(source = taskName, type = type) \
                .orderBy('timestamp'):
                if n > 0:
                    n -= 1

                tb = log.content
                when = log.timestamp

                if not preserve:
                    log.delete()

                yield tb # , when

                if not n:
                    break

    def iterate(self, **criteria):
        from stuphos.db.orm import Log
        from stuphos.db import dbCore

        with dbCore.hubThread(self.dbNamespace):
            return iter(Log.selectBy(**criteria).orderBy('timestamp'))
    __iter__ = iterate

    def deleteAllType(self, type):
        with dbCore.hubThread(self.dbNamespace):
            for e in self.iterate(type = type):
                e.delete(e.id)

    # def waitLogs(self, task, waiter, once = False):
    #     # todo: if already available, call waiter.
    #     # todo: decorator form
    #     self.taskWaiters[task.taskName] = waiter

    def systemLog(self, task, traceback):
        from stuphos.system.api import syslog
        for line in task.formatTraceback(traceback):
            syslog(line)

    @classmethod
    def iterateTracebackData(self, traceback):
        # Compatibility with raise$()d tracebacks that use $stack.
        if traceback is not None:
            try: from ph.emulation.machine.kernel import TracebackFrame, Traceback
            except ImportError:
                Traceback = TracebackFrame = type(None)

            if isinstance(traceback, Traceback):
                # Hide the entire traceback.
                traceback = traceback._iterate()

            for frameTb in traceback:
                if frameTb is None:
                    continue

                if isinstance(frameTb, TracebackFrame):
                    yield (frameTb._frame, frameTb._pos)
                else:
                    yield frameTb


    @staticmethod
    def _formatSource(source):
        return compactify(source, 80).replace('\n', ' ')

    @classmethod
    def buildHasAccess(self, kwd):
        # Build security routines with parameter.
        try: checkAccess = kwd['checkAccess']
        except KeyError:
            def hasAccess(source, frame):
                return True
        else:
            agentCore = runtime[runtime.Agent.System]
            if not agentCore:
                def hasAccess(source, frame):
                    return True
            else:
                def hasAccess(source, frame):
                    if isinstance(source, str):
                        if kwd.get('hide_string_sources'):
                            return

                        # return isYesValue(configuration.Interpreter.reveal_all_string_sources)

                        try: source = frame.locals['path$']
                        except KeyError:
                            try: source = frame.procedure.environ['path$']
                            except (TypeError, AttributeError, KeyError):
                                # The TypeError is for None environs.
                                return # Fail access.


                    # checkAccess (as principal) may be None.
                    source = list(source)
                    # source.append('traceback')

                    return agentCore.principalHasAccess \
                        (checkAccess, source, 'read')

        return hasAccess


    @classmethod
    def formatTraceback(self, traceback, **kwd):
        hasAccess = self.buildHasAccess(kwd)

        for (f, pos) in self.iterateTracebackData(traceback):
            p = f.procedure

            if getattr(p, '_omitTraceback', False):
                continue

            # print(pos, p)

            # Todo: get source code line number from position-to-linenr table,
            # or, yield the Program disassembly of the instruction to string.
            # Note: treating position as if it's the next pc to execute.

            if isinstance(p, Native): # Or SubNative
                # XXX
                try: name = p._native.__name__
                except: pass # Could be a NoAccessException if the native is an Instance.
                else: yield '  %s:' % (name,)

                continue

            try: name = p.name
            except AttributeError:
                name = ''
            else:
                name = f'{name or ""}{name and ":" or ""}'


            # debugOn() # :traceback:
            try: source = p.lineSourceInfo(pos-1)
            except AttributeError:
                source = None

            if source is not None:
                (lnr, source, code) = source
                if hasAccess(source, f):
                    yield f'[{self._formatSource(source)}:{lnr} ({name}{pos})]'

                    if code:
                        yield '  %s' % code.lstrip() # Assume no one's pumping ws at end.

                else:
                    yield '<hidden> (?)'

            else:
                try: source = p.getSourceMap().source
                except AttributeError: source = None

                if source:
                    yield f'[{self._formatSource(source)} ({name}{pos})]'

                try: (start, g) = p.lineGroup(pos)
                except (TypeError, AttributeError):
                    # No line source info found -- render one instruction string.
                    try: s = p.instructionsString(pos-1, pos, suppress_name = True)
                    except AttributeError:
                        s = '<unknown procedure>'
                else:
                    span = 1 # len(g) # XXX :skip: 20 instructions per line??
                    if pos >= start and pos <= (start + span):
                        s = p.instructionsString(start, start + span,
                                                 suppress_name = True,
                                                 highlight = pos)
                    else:
                        # Why lineGroup isn't returning pos-containing span?
                        s = p.instructionsString(pos-1, pos, suppress_name = True)

                n = getattr(p, 'name', None)
                if n:
                    yield f'  {n}'

                yield indent(s.strip())


    # XXX Duplicate code:
    # Todo: rename to formatTracebackContext
    @classmethod
    def formatTracebackHtml(self, traceback, **kwd):
        hasAccess = self.buildHasAccess(kwd)

        for (f, pos) in self.iterateTracebackData(traceback):
            p = f.procedure

            if getattr(p, '_omitTraceback', False):
                continue

            # print(pos, p)

            # Todo: get source code line number from position-to-linenr table,
            # or, yield the Program disassembly of the instruction to string.
            # Note: treating position as if it's the next pc to execute.

            if isinstance(p, Native): # Or SubNative
                # XXX
                try: name = p._native.__name__
                except: pass # Could be a NoAccessException if the native is an Instance.
                else: yield '  %s:' % (name,)

                continue

            try: name = p.name
            except AttributeError:
                name = ''
            else:
                name = f'{name or ""}{name and ":" or ""}'


            # debugOn() # :traceback:
            try: source = p.lineSourceInfo(pos-1)
            except AttributeError:
                source = None

            if source is not None:
                (lnr, source, code) = source
                if hasAccess(source, f):
                    r = dict(sourceFormatted = self._formatSource(source),
                             source = source, lineNr = lnr, name = name,
                             sourceAttr = safeAttr(str(source)),
                             nameAttr = safeAttr(str(name)),
                             position = pos)

                    if code:
                        r['code'] = code

                    yield r

                else:
                    name = '?'
                    yield dict(name = name, nameAttr = safeAttr(str(name)),
                               source = '<hidden>', sourceAttr = '<hidden>',
                               sourceFormatted = '<hidden>', hidden = True)

                    # yield dict(sourceFormatted = self._formatSource(source),
                    #            source = source, name = name,
                    #            sourceAttr = safeAttr(str(source)),
                    #            nameAttr = safeAttr(str(name)))

            else:
                try: source = p.getSourceMap().source
                except AttributeError: source = None

                r = dict()

                if source:
                    r['sourceFormatted'] = self._formatSource(source)
                    r['source'] = source
                    r['name'] = name
                    r['position'] = name

                    r['sourceAttr'] = safeAttr(str(source))
                    r['nameAttr'] = safeAttr(str(name))

                try: (start, g) = p.lineGroup(pos)
                except (TypeError, AttributeError):
                    # No line source info found -- render one instruction string.
                    try: s = p.instructionsString(pos-1, pos, suppress_name = True)
                    except AttributeError:
                        s = '<unknown procedure>'
                else:
                    span = 1 # len(g) # XXX :skip: 20 instructions per line??
                    if pos >= start and pos <= (start + span):
                        s = p.instructionsString(start, start + span,
                                                 suppress_name = True,
                                                 highlight = pos)
                    else:
                        # Why lineGroup isn't returning pos-containing span?
                        s = p.instructionsString(pos-1, pos, suppress_name = True)

                n = getattr(p, 'name', None)
                if n:
                    r['codeName'] = n

                r['code'] = s
                yield r

    formatTracebackContext = formatTracebackHtml


from threading import Thread
from stuphos.runtime.facilities import Facility
from stuphos import getConfig

class ScanThread(Thread, Facility):
    NAME = 'System::Log::Responsive'

    def create(self):
        syslog = getConfig('input-path', 'System:Log')
        syslog = io.path(syslog).open()
        # todo: use the system syslog scanner installed.
        return self(syslog)

    def __init__(self, syslog, scanner):
        self.syslog = syslog
        self.scanner = scanner
        Thread.__init__(self)
        # self.start()

    def run(self):
        while True:
            line = self.syslog.readline()
            self.handle(line, self.scanner.scan(line))

    def handle(self, line, scan):
        pass


try: from stuphos.runtime.registry import registerObject
except ImportError: pass
else:
    class API:
        getScanSyslogPaths = staticmethod(getScanSyslogPaths)

        KNOWN_PATTERN_TYPES = property(lambda self:KNOWN_PATTERN_TYPES)
        compile_pattern = property(lambda self:compile_pattern)
        static_pattern = property(lambda self:static_pattern)

    # Singleton.
    API = API()

    SYSLOG_SCANNER_OBJECT_NAME = 'Syslog::Scanner::API'
    registerObject(SYSLOG_SCANNER_OBJECT_NAME, API)



from django.urls import include, re_path

try: from django.urls import URLPattern as RegexURLPattern, URLResolver as RegexURLResolver
except ImportError:
    from django.urls import RegexURLPattern, RegexURLResolver


# URL pattern objects
def patterns(name, *sequence):
    return list(sequence)

#from django.conf.urls import *



# class unpackableMatch:
#     # Allow unpacking of 
#     class pattern:
#         def __init__(self, *args, **kwd):
#             self.re = reCompile(*args, **kwd)

#         def match(self, *args, **kwd):
#             r = self.re.match(*args, **kwd)
#             if r is None:
#                 return r

#             return unpackableMatch(r)

#     def __init__(self, match):
#         self.match = match

#     def __iter__(self):
#         yield from iter(self.match.groups())


# def url(regex, view, default_args = None, name = None):
#     return RegexURLPattern(reCompile(regex), view, default_args, name)

url = re_path


# def include_urlpatterns(regex, urlpatterns):
#     return RegexURLResolver(reCompile(regex), urlpatterns, None)

class urlConfModule:
    def __init__(self, patterns):
        self.urlpatterns = patterns

def include_urlpatterns(regexp, urlpatterns):
    return re_path(regexp, include(urlConfModule(urlpatterns)))




# Plugin Component Services Manager.
# ---
# Purpose: configuring optional (external to core) components generically
#
from .facilities import Facility, LoadFacility
from .registry import getObject
from . import LookupObject as LookupPluginAction
from stuphos.management.config import loadConfig

from stuphos.etc.tools.logs import logException

from os.path import abspath, exists as fileExists
import traceback

import re
COMPONENT_PATTERN = re.compile('^component(?:\.\d+)?$')
GROUP_PATTERN = re.compile('component-group(?:\.\d+)?$')
FACILITY_PATTERN = re.compile('facility(?:\..+)?$')

PKGDEP_PATTERN = re.compile('package(?:\.\d+)?$')
SVCDEP_PATTERN = re.compile('service(?:\.\d+)?$')

import sys
if sys.platform == 'win32':
    PATH_PATTERN = re.compile('^windows-path(?:\.\d+)?$')
elif sys.platform == 'cygwin':
    PATH_PATTERN = re.compile('^path(?:\.\d+)?$')
else:
    PATH_PATTERN = re.compile('^path(?:\.\d+)?$')

class PluginManager(Facility):
    NAME = 'Plugin::Manager'
    CACHE_NAME = 'Plugin::Manager::Cache'

    def loadBootPlugins(self):
        pluginCache = getObject(self.CACHE_NAME, create = dict)
        thesePlugins = []
        facilities = []

        from stuphos import getSection
        cfg = getSection('Services')
        for opt in cfg.options():
            if COMPONENT_PATTERN.match(opt) is not None:
                plugin = self.loadPluginFromConfig(cfg[opt])
                if plugin is not None:
                    if plugin.name not in pluginCache:
                        pluginCache[plugin.name] = plugin
                        thesePlugins.append(plugin)

            elif GROUP_PATTERN.match(opt) is not None:
                # todo: load a group of services/plugins from file/directory
                pass

            elif FACILITY_PATTERN.match(opt) is not None:
                facilities.append((opt, cfg[opt]))

        # Load facilities from command-line.
        facilities.extend(self.parseServicesCmdline(core.cmdln['options'].services))

        # from ph.interpreter.mental.library.theta import psOpGameVsz
        # print(f'[vsz] {psOpGameVsz()}')

        for (n, f) in sorted(dict(facilities).items(),
                             key = lambda a_b: a_b[0]):

            try: LoadFacility(f)
            except:
                print('Loading [%s]: %s' % (n, f))
                logException(traceback = True)

            # print(f'[vsz] {psOpGameVsz()}')

        for plugin in thesePlugins:
            try: plugin.boot()
            except: traceback.print_exc() # XXX warn :skip:

    __init__ = loadBootPlugins

    def loadPluginFromConfig(self, config_file):
        config_file = abspath(config_file)
        if fileExists(config_file):
            try: return Plugin.LoadFromConfig(config_file)
            except: traceback.print_exc()

    def parseServicesCmdline(self, cmdline):
        for (n, line) in enumerate(cmdline):
            i = line.find('=')
            if i >= 0:
                yield (line[:i], line[i+1:])
            else:
                yield (n, line)


class Plugin:
    class Type(str):
        def __repr__(self):
            return self

    UnknownType = Type('unknown')
    PythonPluginType = Type('python-service')

    @classmethod
    def LoadFromConfig(self, config_file):
        cfg = loadConfig(config_file)
        return self(config_file, cfg)

    MODE_BOOT = 'boot'
    MODE_LOAD = 'load'
    MODE_UNLOAD = 'unload'
    MODE_SHUTDOWN = 'shutdown'

    LIFECYCLE_SECTIONS = [MODE_BOOT, MODE_LOAD, MODE_UNLOAD, MODE_SHUTDOWN]

    def __init__(self, config_file, cfg):
        self.config_file = config_file
        self.mode = None

        # XXX These don't validation as non-None! :skip:
        module = cfg['Module']
        self.name = module['name']
        self.type = module.get('type', self.UnknownType)
        self.package = module['package']
        self.added_paths = []
        self.paths = self.loadPathsFromConfig(module)
        self.lifecycle = {}

        self.dependencies = self.loadDependenciesFromConfig(cfg['Dependencies'])

        for section in cfg.sections():
            mode = section.lower()
            if mode in self.LIFECYCLE_SECTIONS:
                self.lifecycle[mode] = self.loadLifecycle(cfg[section])

    def __repr__(self):
        return '<%s: %s (%s) [%s] %s>' % (self.__class__.__name__, self.name,
                                          self.type, self.package, self.mode)

    class Lifecycle:
        ACTION_NAMES = ['method', 'action', 'object']
        def __init__(self, cfg):
            for name in self.ACTION_NAMES:
                try:
                    value = cfg[name]
                    if value is None:
                        continue

                    self.action = value
                    self.name = name
                    break
                except KeyError:
                    pass
            else:
                raise ValueError('No action name specified in the lifecycle! %s' % self.ACTION_NAMES)

        def __repr__(self):
            return '<%s: action = %s>' % (self.__class__.__name__, self.action)

    loadLifecycle = Lifecycle

    def loadPathsFromConfig(self, module):
        path_collection = []
        for opt in module.options():
            if PATH_PATTERN.match(opt):
                path = module[opt]
                if path and path not in path_collection:
                    path_collection.append(path)

        return path_collection

    def loadPathsIntoSystem(self, these_paths):
        from sys import path as system_path
        added_paths = []
        for path in these_paths:
            path = abspath(path)
            if path not in system_path:
                added_paths.append(path)
                system_path.append(path)

        # For future unload..?
        self.added_paths = added_paths

    def loadDependenciesFromConfig(self, cfg):
        pkgdeps = []
        svcdeps = []

        for opt in cfg.options():
            if PKGDEP_PATTERN.match(opt):
                pkg = cfg.get(opt)
                if pkg not in pkgdeps:
                    pkgdeps.append(pkg)

            elif SVCDEP_PATTERN.match(opt):
               svc = cfg.get(opt)
               if svc not in svcdeps:
                   svcdeps.append(svcdeps)

        deps = []
        for pkg in pkgdeps:
            deps.append(self.PackageDependency(pkg))
        for svc in svcdeps:
            deps.append(self.ServiceDependency(svc))

        return deps

    class PackageDependency:
        def __init__(self, package_name):
            self.package_name = package_name
        def load(self):
            # is this sufficient for full import?
            try: return __import__(self.package_name)
            except ImportError as e:
                from stuphos.system.api import syslog
                syslog('PACKAGE-DEPENDENCY: %s: %s' % (self.package_name, e))
                return False

            return True

    class ServiceDependency:
        def __init__(self, service_name):
            self.service_name = service_name
        def load(self):
            from stuphos.system.api import syslog
            e = NotImplementedError('No way to load service dependencies, yet!')
            syslog('SERVICE-DEPENDENCY: %s: %s' % (self.service_name, e))
            return True

    def loadDependenciesIntoSystem(self, dependencies):
        for dep in dependencies:
            try:
                if not dep.load():
                    return False

            except:
                traceback.print_exc() # todo: log
                return False

        return True

    def lookupAction(self, name):
        if self.type == self.PythonPluginType:
            return LookupPluginAction('%s.%s' % (self.package, name))

        raise TypeError(self.type)

    # Lifecycles.
    def changeMode(self, mode):
        self.mode = mode

        try: action = self.lifecycle[mode].action
        except KeyError: pass
        else:
            action = self.lookupAction(action)
            if callable(action):
                try: action(self, mode)
                except: traceback.print_exc() # XXX change into error mode :skip:

    def boot(self):
        self.loadPathsIntoSystem(self.paths)
        if self.loadDependenciesIntoSystem(self.dependencies):
            self.changeMode(self.MODE_BOOT)
        else:
            # Rollback paths load-in??
            pass

def installPlugins():
    # This is called during mud.bootStart, but eventually I want to be
    # able to hot-plug plugin components, too.
    PluginManager.manage()
    PluginManager.get(create = True)



# Should this be in mud.tasks.parallel?
from _thread import start_new_thread
def nth(function, *args, **kwd):
	return start_new_thread(function, args, kwd)

def nthWith(*args, **kwd):
	# Decorator
	def runWith(function):
		return nth(function, *args, **kwd)

	return runWith



# from stuphmud.server.zones.specials import Special, Register, Any
from stuphmud.server.zones.specials import parseSpecialCommand
from stuphmud.server.player import PlayerAlert
from . import isCommandReserved

from shlex import split as splitShellTokens

class NoOptions(dict):
    def __init__(self):
        self.__dict__ = self
    def __missing__(self, name):
        return False

# Formalization.
##    class SpecialProcedure:
##        # Interface (Abstract).
##        def doSpecial(self, actor, cmd, argstr):
##            SubsystemError('SpecialProcedure', NotImplementedError)
##        def __call__(self, *args, **kwd):
##            return self.doSpecial(*args, **kwd)

##    #@Special('Generic', Any)
##    def genericSpecial(this, actor, command, argstr):
##        pass
##
##    #@Register('Spacecraft::Telepad', Any)
##    class Telepad(Special):
##        room = 3001
##
##        def doLook(self, this, actor, *args):
##            pass

##    class Special(object):
##        def __new__(self, object, name = None, type = Any,
##                    room = None, mobile = None, item = None):
##
##            # Act as decorator, OR instance class.
##            pass
##
##        def __init__(self, name, type = Any):
##            self.name = name
##            self.type = type
##
##        def __lshift__(self, entity):
##            # Todo: validate entity type.
##            entity.special = self
##
##        # XXX lifecycle
##        ##    Assign = Assign
##        ##    Register = Register
##
##        def __call__(self, this, actor, command, argstr):
##            # generic processing:
##            # if isCommandReserved(command):
##            #     return self.periodic(this)
##            # method = getattr(self, 'do%s' % command.name.capitalize() None)
##            # if callable(method):
##            #     return method(This, actor, *argstr.split())
##
##            pass
##
##    class EnvironmentalCommandGrammar:
##        class ParsedCommand:
##            object = subject = directObject = indirectObject = \
##                     verb = preposition = None
##
##        def __init__(self, object):
##            pass
##        def parseCommand(self, args):
##            pass

class Special:
    # Generic base.
    SHLEX = False
    TARGET_SELF = False
    OPTIONS = []

    @classmethod
    def BindSpecial(self, entity, **values):
        # Default binder: if using the Special base class, we should be using an instance anyway.
        return self(**values)

    ##    @classmethod
    ##    def getEntityConfig(self, entity):
    ##        # Try to find a configuration for this entity and the special implementation.
    ##        # If loading from zones.core, then it's related to a zone module, which could
    ##        #   have information about where to find this config.
    ##        pass

    class Commands(dict):
        def Extend(self, **others):
            o = self.copy()
            o.update(others)
            return o

    COMMANDS = Commands()

    class ParsedCommand:
        @classmethod
        def Parse(self, cmd, argstr, shlex = False, options = None):
            if argstr is None:
                argstr = ''

            stripped = argstr.strip()
            if shlex:
                argv = splitShellTokens(argstr)
            else:
                argv = argstr.split()

            if options:
                # Todo: fold some of this into special base.
                # Hickenlooper
                from optparse import OptionParser
                parser = OptionParser()

                for (opt, kwd) in options:
                    parser.add_option(*opt, **kwd)

                (options, argv) = parser.parse_args(argv)
            else:
                options = NoOptions()

            return self(cmd, argstr, argv, stripped, options)

        def __init__(self, cmd, argstr, argv, stripped, options):
            self.command = cmd
            self.argstr = argstr
            self.args = argv
            self.stripped = stripped
            self.options = options
            self.next = iter(self).__next__

            # Presumably, these things don't change.
            self.first = self.first
            self.rest = self.rest

        def isPeriodic(self):
            return isCommandReserved(self.command)
        periodic = property(isPeriodic)

        @property
        def name(self):
            return self.command.name

        @property
        def first(self):
            return self.args[0] if self.args else None

        @property
        def trailing(self):
            (_, trailing) = parseSpecialCommand(self.argstr)
            return trailing

        @property
        def rest(self):
            return self.args[1:]

        @property
        def hasArguments(self):
            return bool(self.args)

        def __iter__(self):
            return iter(self.args)

    def parseCommand(self, cmd, argstr):
        return self.ParsedCommand.Parse(cmd, argstr, shlex = self.SHLEX)

    def findCommandAction(self, cmd, argstr):
        try: return getattr(self, self.COMMANDS[cmd.name])
        except KeyError: pass

    # Match with framework\\game\\world\\player\\__init__.py and structs.h:SPECIAL
    def __call__(self, actor, this, cmd, argstr):
        try:
            action = self.findCommandAction(cmd, argstr)
            if callable(action):
                parsed = self.parseCommand(cmd, argstr)
                if self.TARGET_SELF:
                    first = parsed.first
                    if not (first and actor.find(first) is this):
                        return False

                try:
                    # This re-translates the entity-forward parameter to
                    # be one that follows the actor.
                    if self.dispatchCommand(action, actor,
                                            this, parsed) is False:
                        return False

                except PlayerAlert as e:
                    if actor.peer:
                        e.deliver(actor.peer)

                return True
        except:
            from ph import logException
            logException(traceback = True)

    def dispatchCommand(self, action, *args, **kwd):
        return action(*args, **kwd)



'Test Application'
from ctypes import *
from . import structs
import sys

_ctypes = CDLL({'win32': '_ctypes.pyd',
                'cygwin': 'python2.5/lib-dynload/_ctypes'}[sys.platform])

def make_char():
    return structs.CharacterData()

def make_string(o):
    # Creates a mutable copy.
    return string_at(addressof(o), sizeof(o))

class LibrarySymbol:
    class FrozenTable(Structure):
         _fields_ = [("name", c_char_p),
                     ("code", POINTER(c_ubyte)),
                     ("size", c_int)]

    FrozenTable.P = POINTER(FrozenTable)
    FrozenTable.table = FrozenTable.P.in_dll(pythonapi, "PyImport_FrozenModules")

    CHOP_LENGTH = 10

    @classmethod
    def show_frozen_table(cls):
        print("%-27s   [  Address/Size ]  'Code'" % 'Frozen Module Name')
        for item in cls.FrozenTable.table:
            name = item.name
            if name is None:
                break

            size = item.size
            if size < 0:
                size = -size # abs
                check = '*'
            else:
                check = ' '

            # How to obtain the integer value of the code variable?
            addr = frag = item.code
            addr = 0

            if frag > 0 and False:
                frag = string_at(addr, min(item.size, cls.CHOP_LENGTH))

            print("  %-25s  %c[ 0x%08x/%4d ]  %r" % (name, check, addr, size, frag))

def main():
    global c, s, b

    c = make_char()
    s = make_string(c)
    b = buffer(c)

    # LibrarySymbol.show_frozen_table()

def test():
    from _ctypes import PyObj_FromPtr as fp

    CD = structs.CharacterData
    CD.P = POINTER(CD)
    z = sizeof(CD)

    b = create_string_buffer(z)
    a = addressof(b)

    # return CD.P.from_address(a)

    return CD, CD.P, b, a, z
    print(b, a)

    b2 = (c_char * z)
    b2 = b2(*['x'] * z)

    a = addressof(b2)
    print(b2, a)

    global p, m

    # Decide pointer original value.
    # p = fp(a) # b2
    p = b # 2

    # Perform type-conversion via a cast.
    p = cast(p, CD.P)
    m = p[0]

    m.player.name = 'Fraun'
    print(m.player.name)

class C:
    class Informative:
        def __init__(self, __name, **kwd):
            self.__dict__.update(kwd)
            self.__name = __name
            self.__keys = list(kwd.keys())

        def __str__(self):
            parts = ['[%s]' % self.__name]
            parts.extend('%-30s : %r' % (k, getattr(self, k, None)) \
                         for k in self.__keys)

            return '\n\t'.join(parts)

        __repr__ = __str__

    StructType = type(Structure)
    v = locals()

    from . import structs
    g = structs.__dict__

    for (name, cls) in g.items():
        if type(cls) is StructType:
            try: f = getattr(cls, '_fields_')
            except AttributeError:
                continue

            v[name] = Informative(name, **dict(f))

    del StructType, v, g, name, cls, f

## Test Application -- to be removed.
class CStruct:
    # Create an informative list.
    StructType = type(Structure)
    Assoc = dict()

    # Populate the list.
    for (name, cls) in globals().items():
        if type(cls) is StructType:
            Assoc[name] = cls

    # Cleanup.
    del name, cls

def test():
    Class = MUDOperations
    PointerType = type(POINTER(c_int))
    Primitives = (c_int, c_long, c_char_p, c_size_t) # c_void_p

    # Filter out all func-ptr fields with primitive
    # return types or pointers to primitive types.
    for (name, field) in Class._fields_:
        field = getattr(Class, name)
        field = field.__get__(field)
        restype = field.restype

        if restype in Primitives:
            continue
        if type(restype) is PointerType:
            if restype._type_ in Primitives:
                continue

        print('%s: %s' % (name, restype))

def test():
    from ctypes import Structure, POINTER
    class I(Structure):
        _fields_ = [('x', c_int)]

    ##    @memory
    ##    def Callback():
    ##        return I(42)

    procdef = AUTOFUNCTYPE(POINTER(I))
    proc = procdef(memory(lambda: I(42)))

    print(proc()[0].x)
    return proc


# This is example functionality not necessary in the final product.
def memory(function):
    'Automatically stores reference to ctypes value and returns its address.'
    # (Which is probably what you want.)

    from ctypes import addressof
    def wrapper(*args, **kwd):
        # Expect a ctypes data type.
        value = function(*args, **kwd)
        function.memory = value

        ##    try: memory = function.memory
        ##    except AttributeError:
        ##        memory = function.memory = list()
        ##
        ##    # Can't use set() because ctypes aren't hashable.
        ##    if value not in memory:
        ##        memory.append(value)

        return addressof(value)

    return wrapper


main = test
(__name__ == '__main__') and main()



from sqlobject import *

class StuphItemRecord(SQLObject):
	vnum            = IntCol(unique = True)

	name            = StringCol()
	shortdescr      = StringCol()
	longdescr       = StringCol()
	actiondescr     = StringCol()

	# itemtype? kind?
	type            = IntCol() # Convert into string/text type name (formencode.validators?)

	extraflags      = IntCol()
	wearflags       = IntCol()
	antiflags       = IntCol()
	weight          = IntCol()
	cost            = IntCol()
	minlevel        = IntCol()
	timer           = IntCol()
	trap            = IntCol()

	vals            = BLOBCol() # convert with validator
	extradesc       = BLOBCol() # convert with validator
	affections      = BLOBCol() # convert with validator

	specproc        = StringCol()

	stuphZoneRecord = ForeignKey('StuphZoneRecord')



# StuphMUD 3.0/4.0 Wilderness System.
from stuphos.runtime.facilities import Facility
from stuphmud.server.player import ACMD

# Each room possesses a grid or a collection of grids (rectangle) that define the wilderness shape.
# Each room may possess a relative exterior dimension.
# Each mobile and item possesses a 'spatial' diagram (instance) for describing their location.

class Shape(list):
	class Grid:
		def __init__(self, start, dimensions):
			(startX, startY) = start
			self.start = (startX, startY)
			(width, height) = dimensions
			self.dimensions = (width, height)

		class Tile:
			def __init__(self, xy, code):
				self.code = code

		def __getitem__(self, xy):
			return self.Tile(xy, self.getTileCode(xy[0], xy[1]))

	def __init__(self, size, *grids):
		list.__init__(self, grids)
		(exteriorWidth, exteriorHeight) = size
		self.exterior = (exteriorWidth, exteriorHeight)

class Terrain(Shape):
	def getDefaultTileCode(self):
		pass

	class Grid(Shape.Grid):
		class Tile(Shape.Grid.Tile):
			pass

		def getDefaultTile(self, x, y):
			pass

class WildernessSystem(Facility):
	NAME = 'MUD::Wilderness'

	class Manager(Facility.Manager):
		VERB_NAME = 'wild*erness'
		MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR

	def __init__(self):
		self.dirOptionAssigns = [ACMD('n*orth')(self.doGoNorth),
								 ACMD('s*outh')(self.doGoSouth),
								 ACMD('e*ast')(self.doGoEast),
								 ACMD('w*est')(self.doGoWest),
								 ACMD('u*p')(self.doGoUp),
								 ACMD('d*own')(self.doGoDown)]

		from stuphos import on
		on.lookAtRoom(self.lookAtRoom)

    # Movement -- the Heart of Wilderness.

	def doGoNorth(self, peer, cmd, argstr):
		pass
	def doGoSouth(self, peer, cmd, argstr):
		pass
	def doGoEast(self, peer, cmd, argstr):
		pass
	def doGoWest(self, peer, cmd, argstr):
		pass
	def doGoUp(self, peer, cmd, argstr):
		pass
	def doGoDown(self, peer, cmd, argstr):
		pass

	def lookAtRoom(self, player):
		pass
